#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import pickle
import sys

import pytest

from earthkit.data import config
from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.utils.diag import field_cache_diag

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import load_grib_data  # noqa: E402


class TestMetadataCache:
    def __init__(self):
        self.hits = 0
        self.misses = 0
        self.data = {}

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        self.hits += 1
        return self.data[key]

    def __setitem__(self, key, value):
        self.misses += 1
        self.data[key] = value

    def __len__(self):
        return len(self.data)


@pytest.fixture
def patch_metadata_cache(monkeypatch):
    from earthkit.data.readers.grib.codes import GribField

    def patched_make_metadata_cache(self):
        return TestMetadataCache()

    monkeypatch.setattr(GribField, "_make_metadata_cache", patched_make_metadata_cache)


def _check_diag(diag, ref):
    for k, v in ref.items():
        assert diag[k] == v, f"{k}={diag[k]} != {v}"


@pytest.mark.parametrize("handle_cache_size", [1, 5])
@pytest.mark.parametrize("serialise", [True, False])
def test_grib_cache_basic_file_patched(handle_cache_size, serialise, patch_metadata_cache):

    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": handle_cache_size,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))

        if serialise:
            pickled_f = pickle.dumps(ds)
            ds = pickle.loads(pickled_f)

        assert len(ds) == 18

        # unique values
        ref_vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        # for f in ds:
        #     print(f.metadata()._cache.data)

        diag = ds._cache_diag()
        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": handle_cache_size,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        for i, f in enumerate(ds):
            assert i in ds._field_manager.cache, f"{i} not in cache"
            assert id(f) == id(ds._field_manager.cache[i]), f"{i} not the same object"

        _check_diag(ds._cache_diag(), ref)

        # unique values repeated
        vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert vals == ref_vals

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": handle_cache_size,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_hits": 18 * 4,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        # order by
        ds.order_by(["levelist", "valid_datetime", "paramId", "levtype"])
        diag = ds._cache_diag()
        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": handle_cache_size,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        assert diag["metadata_cache_hits"] >= 18 * 4

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0].handle == md._handle


def test_grib_cache_basic_file_non_patched():
    """This test is the same as test_grib_cache_basic but without the patch_metadata_cache fixture.
    So metadata cache hits and misses are not counted."""

    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ref_vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": 1,
            "handle_create_count": 18,
            "current_handle_count": 0,
            # "metadata_cache_hits": 0,
            # "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        for i, f in enumerate(ds):
            assert i in ds._field_manager.cache, f"{i} not in cache"
            assert id(f) == id(ds._field_manager.cache[i]), f"{i} not the same object"

        _check_diag(ds._cache_diag(), ref)

        # unique values repeated
        vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert vals == ref_vals

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": 1,
            "handle_create_count": 18,
            "current_handle_count": 0,
            # "metadata_cache_hits": 18 * 4,
            # "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        # order by
        ds.order_by(["levelist", "valid_datetime", "paramId", "levtype"])
        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": 1,
            "handle_create_count": 18,
            "current_handle_count": 0,
            # "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0].handle == md._handle


@pytest.mark.parametrize("serialise", [True, False])
@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
def test_grib_cache_basic_metadata_patched(serialise, fl_type, patch_metadata_cache):

    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds, _ = load_grib_data("tuv_pl.grib", fl_type)

        if serialise:
            pickled_f = pickle.dumps(ds)
            ds = pickle.loads(pickled_f)

        assert len(ds) == 18

        # unique values
        ref_vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        # for f in ds:
        #     print(f.metadata()._cache.data)

        diag = ds._cache_diag()
        ref = {
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        # unique values repeated
        vals = ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert vals == ref_vals

        ref = {
            "metadata_cache_hits": 18 * 4,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        # order by
        ds.order_by(["levelist", "valid_datetime", "paramId", "levtype"])
        diag = ds._cache_diag()
        ref = {
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }
        _check_diag(ds._cache_diag(), ref)

        assert diag["metadata_cache_hits"] >= 18 * 4

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        if fl_type != "array":
            # handle is taken from the field
            assert hasattr(md, "_field")
            assert ds[0].handle == md._handle
        else:
            # handle is not taken from the metadata
            assert not hasattr(md, "_field")
            assert ds[0].handle == md._handle


def test_grib_cache_options_1(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "temporary",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert ds._field_manager.cache is not None
        assert ds._handle_manager.cache is None

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_create_count": 18 * 5,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }

        _check_diag(ds._cache_diag(), ref)

        for i, f in enumerate(ds):
            assert i in ds._field_manager.cache, f"{i} not in cache"
            assert id(f) == id(ds._field_manager.cache[i]), f"{i} not the same object"

        _check_diag(ds._cache_diag(), ref)

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle is None

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is None

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        # key already cached
        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        ref["metadata_cache_hits"] += 1
        _check_diag(ds._cache_diag(), ref)

        # uncached key
        first.metadata("level")
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 7, "metadata_cache_size": 7},
        )

        ref["handle_create_count"] += 1
        ref["metadata_cache_misses"] += 1
        ref["metadata_cache_size"] += 1
        _check_diag(ds._cache_diag(), ref)

        assert first.handle != md._handle

        ref["handle_create_count"] += 2
        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_options_2(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "persistent",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert ds._field_manager.cache is not None
        assert ds._handle_manager.cache is None

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_create_count": 18,
            "current_handle_count": 18,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }

        _check_diag(ds._cache_diag(), ref)

        for i, f in enumerate(ds):
            assert i in ds._field_manager.cache, f"{i} not in cache"
            assert id(f) == id(ds._field_manager.cache[i]), f"{i} not the same object"

        _check_diag(ds._cache_diag(), ref)

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle is not None

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is not None
        assert first._handle == md._handle
        assert first.handle == first._handle

        _check_diag(ds._cache_diag(), ref)

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        # key already cached
        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        ref["metadata_cache_hits"] += 1
        _check_diag(ds._cache_diag(), ref)

        # uncached key
        first.metadata("level")
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 7, "metadata_cache_size": 7},
        )

        ref["metadata_cache_misses"] += 1
        ref["metadata_cache_size"] += 1
        _check_diag(ds._cache_diag(), ref)

        assert first.handle == md._handle

        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_options_3(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        assert ds._field_manager.cache is not None
        assert ds._handle_manager.cache is not None

        ref = {
            "field_cache_size": 18,
            "field_create_count": 18,
            "handle_cache_size": 1,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 18 * 6,
            "metadata_cache_size": 18 * 6,
        }

        _check_diag(ds._cache_diag(), ref)

        for i, f in enumerate(ds):
            assert i in ds._field_manager.cache, f"{i} not in cache"
            assert id(f) == id(ds._field_manager.cache[i]), f"{i} not the same object"

        _check_diag(ds._cache_diag(), ref)

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle is None

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is None

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        # key already cached
        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 6, "metadata_cache_size": 6},
        )

        ref["metadata_cache_hits"] += 1
        _check_diag(ds._cache_diag(), ref)

        # uncached key
        first.metadata("level")
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 7, "metadata_cache_size": 7},
        )

        ref["handle_create_count"] += 1
        ref["metadata_cache_misses"] += 1
        ref["metadata_cache_size"] += 1
        _check_diag(ds._cache_diag(), ref)

        assert first.handle == md._handle

        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_options_4(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "temporary",
            "grib-handle-policy": "temporary",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "field_cache_size": 0,
            "field_create_count": 18,
            "handle_cache_size": 0,
            "handle_create_count": 18 * 5,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 0,
            "metadata_cache_size": 0,
        }

        _check_diag(ds._cache_diag(), ref)

        assert ds._field_manager.cache is None
        assert ds._handle_manager.cache is None

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle != md._handle
        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is None
        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        # repeat with indexed field
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_options_5(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "temporary",
            "grib-handle-policy": "persistent",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "field_cache_size": 0,
            "field_create_count": 18,
            "handle_cache_size": 0,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 0,
            "metadata_cache_size": 0,
        }

        _check_diag(ds._cache_diag(), ref)

        assert ds._field_manager.cache is None
        assert ds._handle_manager.cache is None

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle != md._handle
        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is None
        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        assert first._handle is not None

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        assert first._handle is not None

        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        # repeat with indexed field
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )
        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )
        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_options_6(patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "temporary",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "field_cache_size": 0,
            "field_create_count": 18,
            "handle_cache_size": 1,
            "handle_create_count": 18,
            "current_handle_count": 0,
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 0,
            "metadata_cache_size": 0,
        }

        _check_diag(ds._cache_diag(), ref)

        assert ds._field_manager.cache is None
        assert ds._handle_manager.cache is not None

        # metadata object is not decoupled from the field object
        md = ds[0].metadata()
        assert hasattr(md, "_field")
        assert ds[0]._handle != md._handle
        ref["field_create_count"] += 2
        ref["handle_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        # keep a reference to the field
        first = ds[0]
        md = first.metadata()
        assert md._field == first
        assert first._handle is None
        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        first.metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(first),
            {"metadata_cache_hits": 1, "metadata_cache_misses": 1, "metadata_cache_size": 1},
        )

        _check_diag(ds._cache_diag(), ref)

        # repeat with indexed field
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )

        ref["field_create_count"] += 1
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )
        ref["field_create_count"] += 2
        _check_diag(ds._cache_diag(), ref)

        ds[0].metadata("levelist", default=None)
        _check_diag(
            field_cache_diag(ds[0]),
            {"metadata_cache_hits": 0, "metadata_cache_misses": 0, "metadata_cache_size": 0},
        )
        ref["field_create_count"] += 2
        _check_diag(ds._cache_diag(), ref)


def test_grib_cache_file_use_kwargs_1():
    _kwargs = {
        "grib_field_policy": "temporary",
        "grib_handle_policy": "persistent",
        "grib_handle_cache_size": 1,
        "use_grib_metadata_cache": True,
    }

    ds = from_source("file", earthkit_examples_file("tuv_pl.grib"), **_kwargs)
    assert len(ds) == 18

    # unique values
    ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

    ref = {
        "field_cache_size": 0,
        "field_create_count": 18,
        "handle_cache_size": 0,
        "handle_create_count": 18,
        "current_handle_count": 0,
        "metadata_cache_hits": 0,
        "metadata_cache_misses": 0,
        "metadata_cache_size": 0,
    }

    _check_diag(ds._cache_diag(), ref)


def test_grib_cache_file_use_kwargs_2():
    _kwargs = {
        "grib-field-policy": "temporary",
        "grib_handle_policy": "persistent",
        "grib_handle_cache_size": 1,
        "use_grib_metadata_cache": True,
    }

    with pytest.raises(KeyError):
        from_source("file", earthkit_examples_file("tuv_pl.grib"), **_kwargs)


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
def test_grib_cache_metadata_use_kwargs_1(fl_type, patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": False,
        }
    ):

        _kwargs = {
            "use_grib_metadata_cache": True,
        }

        ds, _ = load_grib_data("tuv_pl.grib", fl_type, **_kwargs)

        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 108,
            "metadata_cache_size": 108,
        }

        _check_diag(ds._cache_diag(), ref)

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "metadata_cache_hits": 72,
            "metadata_cache_misses": 108,
            "metadata_cache_size": 108,
        }

        _check_diag(ds._cache_diag(), ref)


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
def test_grib_cache_metadata_use_kwargs_2(fl_type, patch_metadata_cache):
    with config.temporary(
        {
            "grib-field-policy": "persistent",
            "grib-handle-policy": "cache",
            "grib-handle-cache-size": 1,
            "use-grib-metadata-cache": True,
        }
    ):

        _kwargs = {
            "use_grib_metadata_cache": False,
        }

        ds, _ = load_grib_data("tuv_pl.grib", fl_type, **_kwargs)

        assert len(ds) == 18

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 0,
            "metadata_cache_size": 0,
        }

        _check_diag(ds._cache_diag(), ref)

        # unique values
        ds.unique_values("paramId", "levelist", "levtype", "valid_datetime")

        ref = {
            "metadata_cache_hits": 0,
            "metadata_cache_misses": 0,
            "metadata_cache_size": 0,
        }

        _check_diag(ds._cache_diag(), ref)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize(
    "params,expected_meta,metadata_keys",
    [
        (dict(dataCategory=2), [[2]] * 10, []),
        (dict(dataCategory=2, ident="01400"), [[2, "01400"]], []),
        (
            dict(dataCategory=2, ident=["01400", "11747"]),
            [
                [2, "01400"],
                [2, "11747"],
            ],
            [],
        ),
        (dict(dataCategory=22), [], []),
        (dict(INVALIDKEY="w"), [], []),
        (
            dict(dataCategory=[2], ident=["01400", "11747"], dataSubCategory=101),
            [
                [2, "01400", 101],
                [2, "11747", 101],
            ],
            [],
        ),
    ],
)
def test_bufr_sel_single_file_1(params, expected_meta, metadata_keys):
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    g = f.sel(**params)
    assert len(g) == len(expected_meta)
    if len(expected_meta) > 0:
        keys = list(params.keys())
        if metadata_keys:
            keys = metadata_keys

        assert g.metadata(keys) == expected_meta
    return


def test_bufr_sel_single_file_as_dict():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))
    g = f.sel({"dataCategory": 2, "ident": ["01400", "11747"], "dataSubCategory": 101})
    assert len(g) == 2
    assert g.metadata(["dataCategory", "ident", "dataSubCategory"]) == [
        [2, "01400", 101],
        [2, "11747", 101],
    ]


def test_bufr_sel_multi_file():
    f1 = from_source("file", earthkit_examples_file("temp_10.bufr"))
    f2 = from_source("file", earthkit_examples_file("synop_10.bufr"))
    f = from_source("multi", [f1, f2])

    # single resulting message
    g = f.sel(dataCategory=0, ident="68267")
    assert len(g) == 1
    assert g.metadata(["dataCategory", "ident:s"]) == [[0, "68267"]]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import write_to_file

LOG = logging.getLogger(__name__)


def test_multi_graph_1():
    a11 = from_source("dummy-source", kind="grib", date=20000101)
    a12 = from_source("dummy-source", kind="grib", date=20000102)
    b11 = from_source("dummy-source", kind="grib", date=20000103)
    b12 = from_source("dummy-source", kind="grib", date=20000104)

    a21 = from_source("dummy-source", kind="grib", date=20000105)
    a22 = from_source("dummy-source", kind="grib", date=20000106)
    b21 = from_source("dummy-source", kind="grib", date=20000107)
    b22 = from_source("dummy-source", kind="grib", date=20000108)

    m1 = from_source(
        "multi",
        from_source("multi", a11, a12),
        from_source("multi", b11, b12),
    )

    m2 = from_source(
        "multi",
        from_source("multi", a21, a22),
        from_source("multi", b21, b22),
    )

    ds = from_source("multi", m1, m2)
    # ds.graph()

    assert len(ds) == 8
    ds.to_xarray()


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_multi_graph_2(write_method):
    with temp_directory() as tmpdir:
        os.mkdir(os.path.join(tmpdir, "a1"))
        a11 = from_source("dummy-source", kind="grib", date=20000101)
        write_to_file(write_method, os.path.join(tmpdir, "a1", "a11.grib"), a11)
        a12 = from_source("dummy-source", kind="grib", date=20000102)
        write_to_file(write_method, os.path.join(tmpdir, "a1", "a12.grib"), a12)

        os.mkdir(os.path.join(tmpdir, "b1"))
        b11 = from_source("dummy-source", kind="grib", date=20000103)
        write_to_file(write_method, os.path.join(tmpdir, "b1", "b11.grib"), b11)
        b12 = from_source("dummy-source", kind="grib", date=20000104)
        write_to_file(write_method, os.path.join(tmpdir, "b1", "b12.grib"), b12)

        os.mkdir(os.path.join(tmpdir, "a2"))
        a21 = from_source("dummy-source", kind="grib", date=20000105)
        write_to_file(write_method, os.path.join(tmpdir, "a2", "a21.grib"), a21)
        a22 = from_source("dummy-source", kind="grib", date=20000106)
        write_to_file(write_method, os.path.join(tmpdir, "a2", "a22.grib"), a22)

        os.mkdir(os.path.join(tmpdir, "b2"))
        b21 = from_source("dummy-source", kind="grib", date=20000107)
        write_to_file(write_method, os.path.join(tmpdir, "b2", "b21.grib"), b21)
        b22 = from_source("dummy-source", kind="grib", date=20000108)
        write_to_file(write_method, os.path.join(tmpdir, "b2", "b22.grib"), b22)

        def filter(path_or_url):
            return path_or_url.endswith("2.grib")

        ds = from_source("file", tmpdir, filter=filter)
        # ds.graph()

        assert len(ds) == 4


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
@pytest.mark.skipif(sys.platform == "win32", reason="Cannot unlink dir on Windows")
def test_multi_directory_1(write_method):
    with temp_directory() as directory:
        for date in (20000101, 20000102):
            ds = from_source("dummy-source", kind="grib", date=date)
            write_to_file(write_method, os.path.join(directory, f"{date}.grib"), ds)

        ds = from_source("file", directory)
        print(ds)
        assert len(ds) == 2
        ds.graph()

        with temp_file() as filename:
            write_to_file(write_method, filename, ds)
            ds = from_source("file", filename)
            assert len(ds) == 2


def test_multi_grib():
    ds = from_source(
        "multi",
        from_source("dummy-source", kind="grib", date=20000101),
        from_source("dummy-source", kind="grib", date=20000102),
    )
    assert len(ds) == 2
    ds.to_xarray()
    # ds.statistics()


def test_multi_grib_mixed():
    ds = from_source(
        "multi",
        from_source("dummy-source", kind="grib", date=20000101),
        from_source("dummy-source", kind="grib", date=20000102),
        from_source("dummy-source", kind="unknown", hello="world"),
    )
    assert len(ds) == 2


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import mimetypes
import pathlib

from earthkit.data.readers import Reader


class XarrayMixIn:
    def to_xarray(self, **kwargs):
        try:
            from covjsonkit.api import Covjsonkit
        except ImportError:
            raise ImportError("covjson handling requires 'covjsonkit' to be installed")

        decoder = Covjsonkit().decode(self._json())
        return decoder.to_xarray()


class CovjsonReader(XarrayMixIn, Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path})"

    def mutate_source(self):
        # A Covjson is a source itself
        return self

    def _json(self):
        import json

        with open(self.path, "r") as f:
            d = json.load(f)
            return d

    def is_streamable_file(self):
        return True


class CovjsonStreamReader(Reader):
    def __init__(self, stream):
        self._stream = stream

    def __iter__(self):
        return self

    def __next__(self):
        import json

        d = self._stream.read()
        if d:
            return CovjsonInMemory(json.loads(d))
        else:
            raise StopIteration

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def mutate_source(self):
        # A Covjson is a source itself
        return self


class CovjsonMemoryReader(Reader):
    def __init__(self, buf):
        self.buf = buf

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def mutate_source(self):
        import json

        return CovjsonInMemory(json.loads(self.buf))

    @staticmethod
    def _from_stream(stream):
        d = stream.read()
        return CovjsonMemoryReader(d)


class CovjsonInMemory(XarrayMixIn, Reader):
    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def mutate_source(self):
        # A Covjson is a source itself
        return self

    def _json(self):
        return self.data


def _match_content_type(content_type):
    return content_type is not None and content_type == "application/prs.coverage+json"


def _match_magic(magic, deeper_check):
    if magic is not None:
        type_id = b'{"type": "CoverageCollection"'
        if not deeper_check:
            return magic.startswith(type_id)
        else:
            return type_id in magic
    return False


def reader(source, path, *, magic=None, deeper_check=False, content_type=None, **kwargs):
    def _reader():
        return CovjsonReader(source, path)

    if _match_content_type(content_type) or _match_magic(magic, deeper_check):
        return _reader()

    extension = pathlib.Path(path).suffix
    if extension in [".covjson"]:
        return _reader()

    kind, _ = mimetypes.guess_type(path)
    if kind in ["application/prs.cov+json"]:
        return _reader()


def memory_reader(source, buffer, *, magic=None, deeper_check=False, content_type=None, **kwargs):
    if _match_content_type(content_type) or _match_magic(magic, deeper_check):
        return CovjsonMemoryReader(buffer)


def stream_reader(
    source,
    stream,
    *,
    magic=None,
    deeper_check=False,
    content_type=None,
    memory=False,
    **kwargs,
):
    if _match_content_type(content_type) or _match_magic(magic, deeper_check):
        if memory:
            return CovjsonMemoryReader._from_stream(stream)
        return CovjsonStreamReader(stream)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_coords  # noqa: E402
from xr_engine_fixtures import compare_dims  # noqa: E402


def _get_attrs(metadata):
    return {k: metadata.get(k, None) for k in ["gridType", "units"]}


def _get_attrs_for_key_1(key, metadata):
    return str(metadata.get("units", "")) + "_test1"


def _get_attrs_for_key_2(key, metadata):
    return metadata.get(key, "") + "_test2"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,coords,dims,attrs",
    [
        (
            {
                "profile": "mars",
                "dims_as_attrs": None,
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "strict": True,
                "dim_name_from_role_name": False,
            },
            {
                "date": [20240603, 20240604],
                "time": [0, 1200],
                "step_timedelta": [datetime.timedelta(hours=0), datetime.timedelta(hours=6)],
                "levelist": [500, 700],
            },
            {"date": 2, "time": 2, "step_timedelta": 2, "levelist": 2},
            {},
        ),
        (
            {
                "profile": "mars",
                "dims_as_attrs": "levtype",
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "strict": True,
                "dim_name_from_role_name": False,
            },
            {
                "date": [20240603, 20240604],
                "time": [0, 1200],
                "step_timedelta": [datetime.timedelta(hours=0), datetime.timedelta(hours=6)],
                "levelist": [500, 700],
            },
            {"date": 2, "time": 2, "step_timedelta": 2, "levelist": 2},
            {"levtype": 2},
        ),
        (
            {
                "profile": "mars",
                "dims_as_attrs": None,
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "strict": False,
                "dim_name_from_role_name": False,
            },
            {
                "date": [20240603, 20240604],
                "time": [0, 1200],
                "step_timedelta": [datetime.timedelta(hours=0), datetime.timedelta(hours=6)],
                "levelist": [500, 700],
            },
            {"date": 2, "time": 2, "step_timedelta": 2, "levelist": 2},
            {},
        ),
        (
            {
                "profile": "mars",
                "dims_as_attrs": "levtype",
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "strict": False,
                "dim_name_from_role_name": False,
            },
            {
                "date": [20240603, 20240604],
                "time": [0, 1200],
                "step_timedelta": [datetime.timedelta(hours=0), datetime.timedelta(hours=6)],
                "levelist": [500, 700],
            },
            {"date": 2, "time": 2, "step_timedelta": 2, "levelist": 2},
            {"levtype": 2},
        ),
    ],
)
def test_xr_dims_as_attrs(kwargs, coords, dims, attrs):
    ds0 = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl_small.grib")
    )

    ds = ds0.to_xarray(**kwargs)
    compare_coords(ds, coords)
    compare_dims(ds, dims, sizes=True)

    for k, v in attrs.items():
        ds["t"].attrs[k] == v


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,coords,dims,attrs",
    [
        (
            {
                "profile": "mars",
                "attrs_mode": "fixed",
                "variable_attrs": [
                    "shortName",
                    "key=levtype",
                    "namespace=parameter",
                    _get_attrs,
                    {
                        "edition": _get_attrs_for_key_1,
                        "typeOfLevel": _get_attrs_for_key_2,
                        "param": "test",
                        "mykey1": "key=levelist",
                        "mykey2": "namespace=vertical",
                    },
                ],
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "strict": False,
                "dim_name_from_role_name": False,
            },
            {
                "date": [20240603, 20240604],
                "time": [0, 1200],
                "step_timedelta": [datetime.timedelta(hours=0), datetime.timedelta(hours=6)],
                "levelist": [500, 700],
            },
            {"date": 2, "time": 2, "step_timedelta": 2, "levelist": 2},
            {
                "shortName": "t",
                "levtype": "pl",
                "parameter": {
                    "centre": "ecmf",
                    "paramId": 130,
                    "units": "K",
                    "name": "Temperature",
                    "shortName": "t",
                },
                "gridType": "regular_ll",
                "units": "K",
                "edition": "K_test1",
                "typeOfLevel": "isobaricInhPa_test2",
                "param": "test",
                "mykey1": 500,
                "mykey2": {"typeOfLevel": "isobaricInhPa", "level": 500},
            },
        ),
    ],
)
def test_xr_attrs_types(kwargs, coords, dims, attrs):
    ds0 = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl_small.grib")
    )

    ds = ds0.to_xarray(**kwargs)
    compare_coords(ds, coords)
    compare_dims(ds, dims, sizes=True)

    print(ds["t"].attrs)
    for k, v in attrs.items():
        assert ds["t"].attrs[k] == v, f"{k}"


@pytest.mark.cache
def test_xr_global_attrs():
    ds_fl = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl_small.grib")
    )
    ds = ds_fl.to_xarray(
        attrs_mode="fixed",
        global_attrs=[
            {"centre_fixed": "_ecmf_"},
            lambda md: {"centre_callable": md.get("centre")},
            "centre",
            "key=centreDescription",
            {"centre_key": "key=centre"},
            {"geography_namespace": "namespace=geography"},
        ],
    )
    ref_global_attrs = {
        "centre_callable": "ecmf",
        "centre": "ecmf",
        "centreDescription": "European Centre for Medium-Range Weather Forecasts",
        "centre_key": "ecmf",
        "geography_namespace": {
            "Ni": 36,
            "Nj": 19,
            "latitudeOfFirstGridPointInDegrees": 90.0,
            "longitudeOfFirstGridPointInDegrees": 0.0,
            "latitudeOfLastGridPointInDegrees": -90.0,
            "longitudeOfLastGridPointInDegrees": 350.0,
            "iScansNegatively": 0,
            "jScansPositively": 0,
            "jPointsAreConsecutive": 0,
            "jDirectionIncrementInDegrees": 10.0,
            "iDirectionIncrementInDegrees": 10.0,
            "gridType": "regular_ll",
        },
        "centre_fixed": "_ecmf_",
    }
    assert ds.attrs == ref_global_attrs



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from . import cf
from . import proj

try:
    import cartopy.crs as ccrs

    NO_CARTOPY = False
except ImportError:
    NO_CARTOPY = True

CARTOPY_WARNING = "This feature requires 'cartopy' to be installed"


DEFAULT_LATLON_PROJ_STRING = (
    "+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 " "+no_defs +type=crs"
)


class Projection:
    @classmethod
    def from_proj_string(cls, proj_string):
        if proj_string is None:
            proj_string = DEFAULT_LATLON_PROJ_STRING

        proj_params = proj.to_dict(proj_string)

        try:
            proj_name = proj_params["proj"]
        except KeyError:
            raise ValueError("proj string missing required parameter 'proj'")

        for cls in _PROJECTIONS:
            if cls.PROJ_NAME == proj_name:
                break
        else:
            raise ValueError(f"proj projection '{proj_name}' is not supported")

        kwargs = proj.to_projection_kwargs(proj_params)

        return cls(proj_string, **kwargs)

    @classmethod
    def from_cf_grid_mapping(cls, grid_mapping_name, **parameters):
        proj_string = parameters.pop("proj4_params", None)
        if proj_string is not None:
            return Projection.from_proj_string(proj_string)

        for cls in _PROJECTIONS:
            if cls.CF_GRID_MAPPING_NAME == grid_mapping_name:
                break
        else:
            raise ValueError(f"grid mapping '{grid_mapping_name}' is not supported")

        kwargs = cf.to_projection_kwargs(parameters)

        return cls(proj_string=proj_string, **kwargs)

    def __init__(self, proj_string=None, **kwargs):
        self.parameters = kwargs
        self.globe = self.parameters.pop("globe", dict())
        self._proj_string = proj_string

    def __repr__(self):
        if not NO_CARTOPY:
            return self.to_cartopy_crs().__repr__()
        else:
            return self.__str__()

    def __str__(self):
        return self.to_proj_string()

    def to_proj_string(self):
        if self._proj_string is None:
            raise ValueError("projection source provided no proj string")
        return self._proj_string

    def to_cartopy_globe(self):
        if NO_CARTOPY:
            raise ImportError(CARTOPY_WARNING)
        return ccrs.Globe(**self.globe)

    def to_cartopy_crs(self):
        if NO_CARTOPY:
            raise ImportError(CARTOPY_WARNING)
        return getattr(ccrs, self.CARTOPY_CRS)(
            globe=self.to_cartopy_globe(),
            **self.parameters,
        )


class EquidistantCylindrical(Projection):
    PROJ_NAME = "eqc"
    CF_GRID_MAPPING_NAME = "latitude_longitude"
    CARTOPY_CRS = "PlateCarree"


class LongLat(Projection):
    PROJ_NAME = "longlat"
    CF_GRID_MAPPING_NAME = "latitude_longitude"
    CARTOPY_CRS = "PlateCarree"


class LambertAzimuthalEqualArea(Projection):
    PROJ_NAME = "laea"
    CF_GRID_MAPPING_NAME = "lambert_azimuthal_equal_area"
    CARTOPY_CRS = "LambertAzimuthalEqualArea"


class LambertConformal(Projection):
    PROJ_NAME = "lcc"
    CF_GRID_MAPPING_NAME = "lambert_conformal_conic"
    CARTOPY_CRS = "LambertConformal"


class AlbersEqualArea(Projection):
    PROJ_NAME = "aea"
    CF_GRID_MAPPING_NAME = "albers_conical_equal_area"
    CARTOPY_CRS = "AlbersEqualArea"


class Mercator(Projection):
    PROJ_NAME = "merc"
    CF_GRID_MAPPING_NAME = "mercator"
    CARTOPY_CRS = "Mercator"


class TransverseMercator(Projection):
    PROJ_NAME = "tmerc"
    CF_GRID_MAPPING_NAME = "transverse_mercator"
    CARTOPY_CRS = "TransverseMercator"


_PROJECTIONS = [
    EquidistantCylindrical,
    LongLat,
    LambertAzimuthalEqualArea,
    LambertConformal,
    AlbersEqualArea,
    Mercator,
    TransverseMercator,
]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.encoders.grib import GribEncoder
from earthkit.data.targets import to_target
from earthkit.data.targets.fdb import FDBTarget
from earthkit.data.testing import NO_FDB
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file

TEST_GRIB_REQUEST = {
    "class": "od",
    "expver": "0001",
    "stream": "oper",
    "date": "20200513",
    "time": 1200,
    "domain": "g",
    "type": "an",
    "levtype": "sfc",
    "step": 0,
    "param": [167, 151],
}


def make_fdb_config(path):
    fdb_schema = earthkit_test_data_file("fdb_schema.txt")
    fdb_dir = path
    os.makedirs(fdb_dir, exist_ok=True)
    config = {
        "type": "local",
        "engine": "toc",
        "schema": fdb_schema,
        "spaces": [{"handler": "Default", "roots": [{"path": fdb_dir}]}],
    }
    return config


@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"encoder": "grib"},
        {"encoder": GribEncoder()},
    ],
)
@pytest.mark.parametrize("direct_call", [True, False])
def test_target_fdb_grib_core(kwargs, direct_call):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_directory() as tmpdir:
        config = make_fdb_config(os.path.join(tmpdir, "_fdb"))

        if direct_call:
            to_target("fdb", config=config, data=ds, **kwargs)
        else:
            ds.to_target("fdb", config=config, **kwargs)

        # target = FDBTarget(config=config)
        # if per_field:
        #     target.write(ds[0])
        #     target.write(ds[1])
        # else:
        #     target.write(ds)

        # target.flush()

        ds1 = from_source("fdb", TEST_GRIB_REQUEST, config=config, stream=False)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)


@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.parametrize("per_field", [True, False])
def test_target_fdb_grib_direct_api(per_field):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    vals_ref = ds.values[:, :4]

    with temp_directory() as tmpdir:
        config = make_fdb_config(os.path.join(tmpdir, "_fdb"))

        target = FDBTarget(config=config)
        if per_field:
            target.write(ds[0])
            target.write(ds[1])
        else:
            target.write(ds)

        target.flush()

        ds1 = from_source("fdb", TEST_GRIB_REQUEST, config=config, stream=False)
        assert len(ds) == len(ds1)
        assert ds1.metadata("shortName") == ["2t", "msl"]
        assert np.allclose(ds1.values[:, :4], vals_ref)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import ABCMeta
from abc import abstractmethod
from itertools import product

LOG = logging.getLogger(__name__)


class Splitter(metaclass=ABCMeta):
    @staticmethod
    def grids(ds):
        v, _ = ds.unique_values(["md5GridSection"])
        return v["md5GridSection"]

    @abstractmethod
    def split(self, ds, profile):
        pass

    @staticmethod
    def make(split_dims, auto_split=False):
        """TODO: auto_split is not implemented"""
        if not auto_split and not split_dims:
            return NoSplitter()
        elif split_dims:
            return DimSplitter(split_dims)
        raise ValueError("Invalid split configuration")


class NoSplitter(Splitter):
    def split(self, ds, profile):
        grids = self.grids(ds)
        if len(grids) != 1:
            raise ValueError(f"Expected one grid, got {len(grids)}")

        dims = profile.dim_keys
        yield dims, ds


class DimSplitter(Splitter):
    def __init__(self, split_dims):
        self.split_dims = split_dims

    def split(self, builder):
        ds_xr, dims = builder.prepare(self.split_dims)

        for x in product(*dims.values()):
            y = dict(zip(dims.keys(), x))
            ds_sel = ds_xr.sel(**y)
            if len(ds_sel) == 0:
                continue
            ds_sort, profile = builder.parse(ds_sel, None)
            if len(ds_sort) == 0:
                raise ValueError(f"No field found for selection={y}")
            yield ds_sort, profile, y



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.metadata import Metadata
from earthkit.data.core.metadata import WrappedMetadata


class SimpleFieldList(FieldList):
    def __init__(self, fields=None):
        self.fields = fields if fields is not None else []

    def append(self, field):
        self.fields.append(field)

    def _getitem(self, n):
        return self.fields[n]

    def __len__(self):
        return len(self.fields)

    def __repr__(self) -> str:
        return f"SimpleFieldList({len(self)})"

    def __getstate__(self) -> dict:
        ret = {}
        ret["_fields"] = self.fields
        return ret

    def __setstate__(self, state: dict):
        self.fields = state.pop("_fields")

    def to_pandas(self, *args, **kwargs):
        # TODO make it generic
        if len(self) > 0:
            if self[0]._metadata.data_format() == "grib":
                from earthkit.data.readers.grib.pandas import PandasMixIn

                class _C(PandasMixIn, SimpleFieldList):
                    pass

                return _C(self.fields).to_pandas(*args, **kwargs)
        else:
            import pandas as pd

            return pd.DataFrame()

    def to_xarray(self, *args, **kwargs):
        # TODO make it generic
        if len(self) > 0:
            if self[0]._metadata.data_format() in ("grib", "dict"):
                from earthkit.data.readers.grib.xarray import XarrayMixIn

                class _C(XarrayMixIn, SimpleFieldList):
                    pass

                return _C(self.fields).to_xarray(*args, **kwargs)
        else:
            import xarray as xr

            return xr.Dataset()

    def mutate_source(self):
        return self

    @classmethod
    def new_mask_index(cls, *args, **kwargs):
        assert len(args) == 2
        fs = args[0]
        indices = list(args[1])
        return cls.from_fields([fs.fields[i] for i in indices])

    @classmethod
    def merge(cls, sources):
        if not all(isinstance(_, SimpleFieldList) for _ in sources):
            raise ValueError("SimpleFieldList can only be merged to another SimpleFieldLists")

        from itertools import chain

        return cls.from_fields(list(chain(*[f for f in sources])))


class WrappedField:
    def __init__(self, field):
        self._field = field

    def __getattr__(self, name):
        return getattr(self._field, name)

    def __repr__(self) -> str:
        return repr(self._field)


class ClonedFieldCore:
    def __init__(self, field, values=None, metadata=None, **kwargs):
        self._field = field
        self.__values = values

        if metadata is not None:
            if isinstance(metadata, dict):
                metadata = dict(**metadata)
            elif isinstance(metadata, (Metadata, WrappedMetadata)):
                if kwargs:
                    raise ValueError("metadata and kwargs cannot be used together")
            else:
                raise ValueError(
                    "metadata must be a dict, Metadata or WrappedMetadata, got %s" % type(metadata)
                )

        if metadata is None:
            metadata = dict()

        assert metadata is not None

        if kwargs:
            if isinstance(metadata, dict):
                metadata.update(kwargs)

        if metadata:
            if isinstance(metadata, dict):
                self.__metadata = WrappedMetadata(field._metadata, extra=metadata, owner=field)
            else:
                self.__metadata = metadata
        else:
            self.__metadata = field._metadata

    def _values(self, dtype=None):
        if self.__values is None:
            return self._field._values(dtype=dtype)
        else:
            if dtype is None:
                return self.__values
            return self.__values.astype(dtype)

    @property
    def _metadata(self):
        return self.__metadata

    def _has_new_values(self):
        return self._values

    @property
    def handle(self):
        return self._metadata._handle

    def _encode(self, encoder, **kwargs):
        """Double dispatch to the encoder"""
        md = {}
        # wrapped metadata
        if hasattr(self._metadata, "extra"):
            md = {k: self._metadata._extra_value(k) for k, v in self._metadata.extra.items()}

        metadata = kwargs.pop("metadata", {})
        metadata.update(md)

        values = kwargs.pop("values", None)
        if values is None:
            values = self._values()

        return encoder._encode_field(self, values=values, metadata=metadata, **kwargs)


# For backwards compatibility
FieldArray = SimpleFieldList



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils.xarray.fieldlist import XArrayInputFieldList


def load_fieldlist(d):
    assert isinstance(d, dict)
    r = []

    prototype = {
        "gridType": "regular_ll",
        "Nx": 2,
        "Ny": 3,
        "distinctLatitudes": [-10.0, 0.0, 10.0],
        "distinctLongitudes": [0.0, 10.0],
        "values": [[1, 2], [3, 4], [5, 6]],
    }

    from itertools import product

    for x in list(product(*d.values())):
        y = dict(zip(d.keys(), x))
        r.append({**y, **prototype})

    return from_source("list-of-dicts", r)


def load_wrapped_fieldlist(d, profile, **kwargs):
    if isinstance(d, list):
        ds = []
        for x in d:
            ds.extend(load_fieldlist(x))
    elif isinstance(d, dict):
        ds = load_fieldlist(d)
    else:
        raise ValueError(f"Invalid data type={type(d)}")

    return XArrayInputFieldList(ds, keys=profile.index_keys, **kwargs)


def compare_dims(ds, ref_coords, order_ref_var=None, sizes=False):
    compare_dim_order(ds, ref_coords, order_ref_var=order_ref_var)
    if not sizes:
        for k, v in ref_coords.items():
            compare_coord(ds, k, v, mode="dim")
    else:
        compare_dim_size(ds, ref_coords)


def compare_coords(ds, ref_coords):
    for k, v in ref_coords.items():
        compare_coord(ds, k, v, mode="coord")


def compare_coord(ds, name, ref_vals, mode="coord"):
    assert name in ds.coords, f"{name=} not in {ds.coords}"
    if mode == "dim":
        assert name in ds.sizes, f"{name=} not in {ds.sizes}"
        assert ds.sizes[name] == len(ref_vals), f"{name=} {ds.sizes[name]} != {len(ref_vals)}"

    vals = ds.coords[name].values
    if isinstance(ref_vals[0], str):
        assert vals.tolist() == ref_vals, f"{name=} {vals.tolist()} != {ref_vals}"
    else:
        vals = np.asarray(vals).flatten()
        ref_vals = np.asarray(ref_vals).flatten()

        assert vals.shape == ref_vals.shape, f"{name=} {vals.shape} != {ref_vals.shape}"
        # datetime/timedelta 64
        if np.issubdtype(vals.dtype, np.datetime64) or np.issubdtype(vals.dtype, np.timedelta64):
            for i in range(len(ref_vals)):
                assert vals[i] == ref_vals[i], f"{name=} {vals[i]} != {ref_vals[i]}"
        # other arrays
        else:
            assert np.allclose(ds.coords[name].values, vals), f"{name=} {ds.coords[name].values} != {vals}"


def compare_dim_order(ds, dims, order_ref_var, check_coord=True):
    if order_ref_var is None:
        return

    dim_order = []

    for d in ds[order_ref_var].dims:
        if d in dims:
            dim_order.append(d)
            if check_coord:
                assert d in ds.coords, f"{d} not in {ds.coords}"

    if isinstance(dims, dict):
        assert dim_order == list(dims.keys()), f"{dim_order=} != {list(dims.keys())}"
    else:
        assert dim_order == dims, f"{dim_order=} != {dims}"


def compare_dim_size(ds, dims):
    for name, v in dims.items():
        assert name in ds.sizes, f"{name=} not in {ds.sizes}"
        assert ds.sizes[name] == v, f"{name=} {ds.sizes[name]} != {v}"


def load_grib_data(filename, source, stream=False, folder="example"):
    if source == "url":
        path = earthkit_remote_test_data_file(filename)
    else:
        if folder == "example":
            path = earthkit_examples_file(filename)
        elif folder == "data":
            path = earthkit_test_data_file(filename)
        else:
            raise ValueError("Invalid folder={folder}")

    if source == "file":
        ds = from_source("file", path)
    elif source == "url":
        ds = from_source("url", path)
    else:
        raise ValueError("Invalid source={source}")

    ds_ref = ds

    if stream:
        f = open(ds.path, "rb")
        ds = from_source("stream", f)

    return ds, ds_ref



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.targets import to_target
from earthkit.data.testing import NO_ZARR


@pytest.mark.skipif(NO_ZARR, reason="Zarr not installed")
@pytest.mark.cache
@pytest.mark.parametrize("direct_call", [True, False])
def test_target_zarr_from_grib(direct_call):
    ds = from_source("sample", "pl.grib")

    with temp_directory() as tmp:
        path = os.path.join(tmp, "_res.zarr")

        if direct_call:
            to_target(
                "zarr",
                earthkit_to_xarray_kwargs={"chunks": {"forecast_reference_time": 1, "step": 1, "level": 1}},
                xarray_to_zarr_kwargs={"store": path, "mode": "w"},
                data=ds,
            )
        else:
            ds.to_target(
                "zarr",
                earthkit_to_xarray_kwargs={"chunks": {"forecast_reference_time": 1, "step": 1, "level": 1}},
                xarray_to_zarr_kwargs={"store": path, "mode": "w"},
            )

        import zarr

        root = zarr.group(path)
        assert root

        shapes = {
            "t": (4, 2, 2, 19, 36),
            "r": (4, 2, 2, 19, 36),
            "forecast_reference_time": (4,),
            "step": (2,),
            "level": (2,),
            "latitude": (19,),
            "longitude": (36,),
        }

        for k in ["t", "r", "forecast_reference_time", "step", "level", "latitude", "longitude"]:
            k in root, f"Key {k} not found in Zarr root"
            assert (
                root[k].shape == shapes[k]
            ), f"Shape mismatch for {k}: expected {shapes[k]}, got {root[k].shape}"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils.patterns import HivePattern


@pytest.mark.parametrize("fx", ["hive_fs_1", "hive_fs_2", "hive_fs_3", "hive_fs_4", "hive_fs_5"])
def test_hive_full_scan(request, fx):
    pattern, files, _ = request.getfixturevalue(fx)

    p = HivePattern(pattern, {})
    res_files = p.scan()

    assert sorted(files) == sorted(res_files)


@pytest.mark.parametrize("fx", ["hive_fs_5"])
def test_hive_full_scan_with_fixed(request, fx):
    pattern, files, values = request.getfixturevalue(fx)

    v = {
        "shortName": values["shortName"],
    }

    p = HivePattern(pattern, v)
    res_files = p.scan()

    assert sorted(files) == sorted(res_files)


@pytest.mark.parametrize(
    "fx,filters,expected_files",
    [
        (
            "hive_fs_1",
            {"shortName": "z", "step": 6},
            ["/my_root/z_2020-05-11+6.grib", "/my_root/z_2020-05-12+6.grib"],
        ),
        (
            "hive_fs_1",
            {"shortName": ["z", "t"], "step": 6},
            [
                "/my_root/z_2020-05-11+6.grib",
                "/my_root/z_2020-05-12+6.grib",
                "/my_root/t_2020-05-11+6.grib",
                "/my_root/t_2020-05-12+6.grib",
            ],
        ),
        (
            "hive_fs_2",
            {"shortName": "z", "step": 6},
            ["/my_root/z/2020-05-11:6.grib", "/my_root/z/2020-05-12:6.grib"],
        ),
        (
            "hive_fs_2",
            {"shortName": ["z", "t"], "step": 6},
            [
                "/my_root/z/2020-05-11:6.grib",
                "/my_root/z/2020-05-12:6.grib",
                "/my_root/t/2020-05-11:6.grib",
                "/my_root/t/2020-05-12:6.grib",
            ],
        ),
        (
            "hive_fs_3",
            {"shortName": "z", "step": 6},
            ["/my_root/6__/z/my+2020-05-11.grib", "/my_root/6__/z/my+2020-05-12.grib"],
        ),
        (
            "hive_fs_3",
            {"shortName": ["z", "t"], "step": 6},
            [
                "/my_root/6__/z/my+2020-05-11.grib",
                "/my_root/6__/z/my+2020-05-12.grib",
                "/my_root/6__/t/my+2020-05-11.grib",
                "/my_root/6__/t/my+2020-05-12.grib",
            ],
        ),
        (
            "hive_fs_4",
            {"shortName": "z", "step": 6},
            ["/my_root/6__/z/z_2020-05-11.grib", "/my_root/6__/z/z_2020-05-12.grib"],
        ),
        (
            "hive_fs_4",
            {"shortName": ["z", "t"], "step": 6},
            [
                "/my_root/6__/z/z_2020-05-11.grib",
                "/my_root/6__/z/z_2020-05-12.grib",
                "/my_root/6__/t/t_2020-05-11.grib",
                "/my_root/6__/t/t_2020-05-12.grib",
            ],
        ),
    ],
)
def test_hive_filter(request, fx, filters, expected_files):
    pattern, _, _ = request.getfixturevalue(fx)

    # files = _build_fs(md, fs_pattern, date_format)
    # print(f"files={files}")

    p = HivePattern(pattern, {})
    res_files = p.scan(filters)

    # print(f"res_files={res_files}")

    assert sorted(expected_files) == sorted(res_files)


class HiveDiag:
    def __init__(self):
        self.file_count = 0
        self.sel_count = 0

    def file(self, count):
        self.file_count += count

    def sel(self, count):
        self.sel_count += count

    def reset(self):
        self.file_count = 0
        self.sel_count = 0


def test_hive_sel_1():
    root = earthkit_test_data_file("pattern/1")
    pattern = "{shortName}_{date:date(%Y-%m-%dT-H-%M)}_{step}.grib"

    ds = from_source("file-pattern", os.path.join(root, pattern), hive_partitioning=True)

    # assert ds.root == path
    diag = HiveDiag()
    # using hive partitioning keys
    r = ds.sel(shortName="t", step=12, _hive_diag=diag)
    assert diag.file_count == 1
    assert diag.sel_count == 0
    assert len(r) == 6

    md_ref = [("t", 1000, 12), ("t", 850, 12), ("t", 700, 12), ("t", 500, 12), ("t", 400, 12), ("t", 300, 12)]
    assert r.metadata("shortName", "level", "step") == md_ref

    # using hive partitioning keys + extra keys from GRIB header
    diag.reset()
    r = ds.sel(shortName="t", step=12, levtype="pl", _hive_diag=diag)
    assert diag.file_count == 1
    assert diag.sel_count == 1
    assert len(r) == 6


def test_hive_sel_2():
    root = earthkit_test_data_file("pattern/invalid")
    pattern = "_{shortName}_{date:date(%Y-%m-%dT%H:%M)}_{step}.grib"

    ds = from_source("file-pattern", os.path.join(root, pattern), hive_partitioning=True)

    r = ds.sel(shortName="t", step=12)
    assert len(r) == 0


def test_hive_init_1():
    pattern = "{shortName}_{date:date(%Y-%m-%dT%H:%M)}_{step}.grib"
    p = HivePattern(pattern)

    assert p.pattern == pattern
    assert p.params == ["shortName", "date", "step"]
    assert p.dynamic_params == ["shortName", "date", "step"]
    assert p.fixed_single_params == {}
    assert p.fixed_multi_params == {}
    assert p.root == ""
    assert p.rest == pattern

    assert len(p.parts) == 1
    ref = ["shortName", "date", "step"]
    for i, v in enumerate(p.parts[0].variables):
        assert v.name == ref[i]


def test_hive_init_2():
    pattern = "root_d/{year}/fc/t_{level}b_/a{shortName}_{date:date(%Y-%m-%dT%H:%M)}_{step}.grib"
    p = HivePattern(pattern)

    assert p.pattern == pattern
    assert p.params == ["year", "level", "shortName", "date", "step"]
    assert p.dynamic_params == ["year", "level", "shortName", "date", "step"]
    assert p.fixed_single_params == {}
    assert p.fixed_multi_params == {}
    assert p.root == "root_d"
    assert p.rest == pattern[7:]

    assert len(p.parts) == 4
    ref = [["year"], [], ["level"], ["shortName", "date", "step"]]
    for i, part in enumerate(p.parts):
        assert len(part.variables) == len(ref[i])
        for k, v in enumerate(part.variables):
            assert v.name == ref[i][k]

    ref = [False, True, False, False]
    for i, part in enumerate(p.parts):
        assert part.is_constant() == ref[i]

    assert p.parts[0].match("2023")
    assert p.parts[0].match("2_2_3")
    assert p.parts[1].match("fc")
    assert not p.parts[1].match("an")
    assert p.parts[2].match("t_500_b_")
    assert not p.parts[2].match("t_500_b")
    assert not p.parts[2].match("500")

    m = p.parts[3].match("at2m_2023-01-21_345.grib")
    assert m is not None
    assert m.groupdict() == {"shortName": "t2m", "date": "2023-01-21", "step": "345"}

    m = p.parts[3].match("a_t2m_2023-01-21_345.grib")
    assert m is not None
    assert m.groupdict() == {"shortName": "_t2m", "date": "2023-01-21", "step": "345"}

    m = p.parts[3].match("at2m2023-01-21_345.grib")
    assert m is None


def test_hive_init_3():
    pattern = "root_d/{year}/fc/t_{level}b_/a{shortName}_{date:date(%Y-%m-%dT%H:%M)}_{step}.grib"
    p = HivePattern(pattern, {"year": [2023, 2024], "level": "500", "shortName": "t"})

    assert p.pattern == pattern
    assert set(p.params) == {"year", "level", "shortName", "date", "step"}
    assert p.dynamic_params == ["date", "step"]
    assert p.fixed_single_params == {"level": "500", "shortName": "t"}
    assert p.fixed_multi_params == {"year": [2023, 2024]}
    assert p.root == "root_d"
    assert p.rest == "{year}/fc/t_500b_/at_{date:date(%Y-%m-%dT%H:%M)}_{step}.grib"

    assert len(p.parts) == 4
    ref = [["year"], [], [], ["date", "step"]]
    for i, part in enumerate(p.parts):
        assert len(part.variables) == len(ref[i])
        for k, v in enumerate(part.variables):
            assert v.name == ref[i][k]

    ref = [False, True, True, False]
    for i, part in enumerate(p.parts):
        assert part.is_constant() == ref[i]

    assert p.parts[0].match("2023")
    assert p.parts[0].match("2024")
    assert p.parts[0].match("2025")
    assert p.parts[0].match("2_2_3")
    assert p.parts[1].match("fc")
    assert not p.parts[1].match("an")
    assert p.parts[2].match("t_500b_")
    assert not p.parts[2].match("at_500b_")
    assert not p.parts[2].match("t_500b__")
    assert not p.parts[2].match("t_500_b")
    assert not p.parts[2].match("500")

    m = p.parts[3].match("at_2023-01-21_345.grib")
    assert m is not None
    assert m.groupdict() == {"date": "2023-01-21", "step": "345"}

    m = p.parts[3].match("a_t_2023-01-21_345.grib")
    assert m is None

    m = p.parts[3].match("at2023-01-21_345.grib")
    assert m is None



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.sources.stream import StreamFieldList
from earthkit.data.testing import ARRAY_BACKENDS
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import write_to_file


def repeat_list_items(items, count):
    return sum([[x] * count for x in items], [])


# @pytest.mark.parametrize(
#     "_kwargs,error",
#     [
#         # (dict(order_by="level"), TypeError),
#         (dict(group_by=1), TypeError),
#         (dict(group_by=["level", 1]), TypeError),
#         # (dict(group_by="level", batch_size=1), TypeError),
#         (dict(batch_size=-1), ValueError),
#     ],
# )
# def test_grib_from_stream_invalid_args(_kwargs, error):
#     with open(earthkit_examples_file("test6.grib"), "rb") as stream:
#         with pytest.raises(error):
#             from_source("stream", stream, **_kwargs)


def test_grib_from_stream_iter():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source("stream", stream)

        # no fieldlist methods are available
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        ref = [
            ("t", 1000),
            ("u", 1000),
            ("v", 1000),
            ("t", 850),
            ("u", 850),
            ("v", 850),
        ]

        for i, f in enumerate(ds):
            assert f.metadata(("param", "level")) == ref[i], i

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["t"], ["u"], ["v"], ["t"], ["u"], ["v"]]),
        ({"n": 3}, [["t", "u", "v"], ["t", "u", "v"]]),
        ({"n": 4}, [["t", "u", "v", "t"], ["u", "v"]]),
    ],
)
def test_grib_from_stream_batched(_kwargs, expected_meta):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source("stream", stream)

        # no methods are available
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        for i, f in enumerate(ds.batched(_kwargs["n"])):
            assert len(f) == len(expected_meta[i])
            f.metadata("param") == expected_meta[i]

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "convert_kwargs,expected_shape",
    [
        ({}, (2, 7, 12)),
        (None, (2, 7, 12)),
        (None, (2, 7, 12)),
        ({"flatten": False}, (2, 7, 12)),
        (
            {"flatten": True},
            (
                2,
                84,
            ),
        ),
    ],
)
def test_grib_from_stream_batched_convert_to_numpy(convert_kwargs, expected_shape):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source("stream", stream)

        ref = [
            [("t", 1000), ("u", 1000)],
            [("v", 1000), ("t", 850)],
            [("u", 850), ("v", 850)],
        ]

        if convert_kwargs is None:
            convert_kwargs = {}

        for i, f in enumerate(ds.batched(2)):
            df = f.to_fieldlist(array_backend="numpy", **convert_kwargs)
            assert df.metadata(("param", "level")) == ref[i], i
            assert df.to_numpy(**convert_kwargs).shape == expected_shape, i
            df1 = df.to_fieldlist(array_backend="numpy", **convert_kwargs)
            assert df1 is not df, i
            assert df1.metadata(("param", "level")) == ref[i], i
            assert df1.to_numpy(**convert_kwargs).shape == expected_shape, i

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("array_backend", ARRAY_BACKENDS)
@pytest.mark.parametrize("group", ["level", ["level", "gridType"]])
def test_grib_from_stream_group_by(array_backend, group):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source("stream", stream)

        # no methods are available
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        ref = [
            [("t", 1000), ("u", 1000), ("v", 1000)],
            [("t", 850), ("u", 850), ("v", 850)],
        ]
        for i, f in enumerate(ds.group_by(group)):
            assert len(f) == 3
            assert f.metadata(("param", "level")) == ref[i]
            afl = f.to_fieldlist(array_backend=array_backend.name)
            assert afl is not f
            assert len(afl) == 3

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "convert_kwargs,expected_shape",
    [
        ({}, (3, 7, 12)),
        (None, (3, 7, 12)),
        (None, (3, 7, 12)),
        ({"flatten": False}, (3, 7, 12)),
        ({"flatten": True}, (3, 84)),
    ],
)
def test_grib_from_stream_group_by_convert_to_numpy(convert_kwargs, expected_shape):
    group = "level"
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source("stream", stream)

        # no fieldlist methods are available on a StreamSource
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        ref = [
            [("t", 1000), ("u", 1000), ("v", 1000)],
            [("t", 850), ("u", 850), ("v", 850)],
        ]

        if convert_kwargs is None:
            convert_kwargs = {}

        for i, f in enumerate(ds.group_by(group)):
            df = f.to_fieldlist(array_backend="numpy", **convert_kwargs)
            assert len(df) == 3
            assert df.metadata(("param", "level")) == ref[i]
            assert df.to_numpy(**convert_kwargs).shape == expected_shape
            df1 = df.to_fieldlist(array_backend="numpy", **convert_kwargs)
            assert df1 is not df
            assert len(df1) == 3
            assert df1.metadata(("param", "level")) == ref[i]
            assert df1.to_numpy(**convert_kwargs).shape == expected_shape

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


def test_grib_from_stream_in_memory():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds = from_source(
            "stream",
            stream,
            read_all=True,
        )

        assert len(ds) == 6

        expected_shape = (6, 7, 12)
        md_ref = [
            ("t", 1000),
            ("u", 1000),
            ("v", 1000),
            ("t", 850),
            ("u", 850),
            ("v", 850),
        ]

        # iteration
        val = [f.metadata(("param", "level")) for f in ds]
        assert val == md_ref, "iteration"

        # metadata
        val = ds.metadata(("param", "level"))
        assert val == md_ref, "method"

        # data
        assert ds.to_numpy().shape == expected_shape

        ref = np.array(
            [
                272.56417847,
                -6.28688049,
                7.83348083,
                272.53916931,
                -4.89837646,
                8.66096497,
            ]
        )

        vals = ds.to_numpy()[:, 0, 0]
        assert np.allclose(vals, ref)

        # slicing
        r = ds[0:3]
        assert len(r) == 3
        val = r.metadata(("param", "level"))
        assert val == md_ref[0:3]

        r = ds[-2:]
        assert len(r) == 2
        val = r.metadata(("param", "level"))
        assert val == md_ref[-2:]

        r = ds.sel(param="t")
        assert len(r) == 2
        val = r.metadata(("param", "level"))
        assert val == [
            ("t", 1000),
            ("t", 850),
        ]


@pytest.mark.parametrize(
    "convert_kwargs,expected_shape",
    [
        ({}, (6, 7, 12)),
        ({"flatten": False}, (6, 7, 12)),
        ({"flatten": True}, (6, 84)),
    ],
)
def test_grib_from_stream_in_memory_convert_to_numpy(convert_kwargs, expected_shape):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        ds_s = from_source("stream", stream, read_all=True)

        ds = ds_s.to_fieldlist(array_backend="numpy", **convert_kwargs)

        assert len(ds) == 6

        ref = ["t", "u", "v", "t", "u", "v"]

        # iteration
        val = [f.metadata("param") for f in ds]
        assert val == ref, "iteration"

        # metadata
        val = ds.metadata("param")
        assert val == ref, "method"

        # data
        assert ds.to_numpy(**convert_kwargs).shape == expected_shape

        ref = np.array(
            [
                272.56417847,
                -6.28688049,
                7.83348083,
                272.53916931,
                -4.89837646,
                8.66096497,
            ]
        )

        if len(expected_shape) == 3:
            vals = ds.to_numpy(**convert_kwargs)[:, 0, 0]
        else:
            vals = ds.to_numpy(**convert_kwargs)[:, 0]

        assert np.allclose(vals, ref)
        assert ds.to_numpy(**convert_kwargs).shape == expected_shape
        ds1 = ds.to_fieldlist(array_backend="numpy", **convert_kwargs)
        assert ds1 is not ds
        assert len(ds1) == 6
        assert ds1.to_numpy(**convert_kwargs).shape == expected_shape


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_when_loaded_from_stream(write_method):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream, read_all=True)
        assert len(fs) == 6
        with temp_file() as tmp:
            write_to_file(write_method, tmp, fs)
            fs_saved = from_source("file", tmp)
            assert len(fs) == len(fs_saved)


def test_grib_multi_from_stream_iter():
    stream1 = open(earthkit_examples_file("test.grib"), "rb")
    stream2 = open(earthkit_examples_file("test4.grib"), "rb")
    ds = from_source("stream", [stream1, stream2])

    assert isinstance(ds, StreamFieldList)

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
    ]

    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["2t"], ["msl"], ["t"], ["z"], ["t"], ["z"]]),
        ({"n": 2}, [["2t", "msl"], ["t", "z"], ["t", "z"]]),
        ({"n": 3}, [["2t", "msl", "t"], ["z", "t", "z"]]),
        ({"n": 4}, [["2t", "msl", "t", "z"], ["t", "z"]]),
    ],
)
def test_grib_multi_grib_from_stream_batched(_kwargs, expected_meta):
    stream1 = open(earthkit_examples_file("test.grib"), "rb")
    stream2 = open(earthkit_examples_file("test4.grib"), "rb")
    ds = from_source("stream", [stream1, stream2])

    assert isinstance(ds, StreamFieldList)

    # no methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds.batched(_kwargs["n"])):
        assert len(f) == len(expected_meta[i])
        f.metadata("param") == expected_meta[i]
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


def test_grib_multi_stream_memory():
    stream1 = open(earthkit_examples_file("test.grib"), "rb")
    stream2 = open(earthkit_examples_file("test4.grib"), "rb")
    ds = from_source("stream", [stream1, stream2], read_all=True)

    assert len(ds) == 6

    md_ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
    ]
    # iteration
    val = [f.metadata(("param", "level")) for f in ds]
    assert val == md_ref, "iteration"

    # metadata
    val = ds.metadata(("param", "level"))
    assert val == md_ref, "method"

    # data
    with pytest.raises(ValueError):
        ds.to_numpy().shape

    # first part
    expected_shape = (2, 11, 19)
    assert ds[0:2].to_numpy().shape == expected_shape

    ref = np.array([262.78027344, 101947.8125])

    vals = ds[0:2].to_numpy()[:, 0, 0]
    assert np.allclose(vals, ref)

    # second part
    expected_shape = (4, 181, 360)
    assert ds[2:].to_numpy().shape == expected_shape

    ref = np.array([228.04600525, 48126.859375, 246.61032104, 11786.1132812])

    vals = ds[2:].to_numpy()[:, 0, 0]
    assert np.allclose(vals, ref)

    # slicing
    r = ds[0:3]
    assert len(r) == 3
    val = r.metadata(("param", "level"))
    assert val == md_ref[0:3]

    r = ds[-2:]
    assert len(r) == 2
    val = r.metadata(("param", "level"))
    assert val == md_ref[-2:]

    r = ds.sel(param="t")
    assert len(r) == 2
    val = r.metadata(("param", "level"))
    assert val == [
        ("t", 500),
        ("t", 850),
    ]


def test_grib_concat_stream():
    stream1 = open(earthkit_examples_file("test.grib"), "rb")
    ds1 = from_source("stream", stream1)
    ds2 = from_source("file", earthkit_examples_file("test4.grib"), stream=True)
    ds3 = from_source("url", earthkit_remote_test_data_file("examples/test6.grib"), stream=True)

    ds = ds1 + ds2 + ds3

    ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


def test_grib_concat_stream_memory():
    stream1 = open(earthkit_examples_file("test.grib"), "rb")
    ds1 = from_source("stream", stream1, read_all=True)
    ds2 = from_source("file", earthkit_examples_file("test4.grib"), stream=True, read_all=True)
    ds3 = from_source(
        "url", earthkit_remote_test_data_file("examples/test6.grib"), stream=True, read_all=True
    )

    ds = ds1 + ds2 + ds3

    ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]

    assert len(ds) == len(ref)
    assert ds.metadata(("param", "level")) == ref

    # repeat the test to check that data is still in memory
    assert len(ds) == len(ref)
    assert ds.metadata(("param", "level")) == ref


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest
import yaml

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.metadata import RestrictedGribMetadata
from earthkit.data.readers.grib.metadata import StandAloneGribMetadata
from earthkit.data.readers.grib.metadata import WrappedMetadata
from earthkit.data.readers.grib.output import new_grib_output
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils import ensure_iterable


def to_tuple(x):
    return tuple(ensure_iterable(x))


def grid_list(files=None, skip=None):
    with open(earthkit_test_data_file(os.path.join("xr_engine", "xr_grid.yaml")), "r") as f:
        r = yaml.safe_load(f)

    files = [] if files is None else files
    skip = [] if skip is None else skip
    for item in r:
        if not files or item["file"] in files and item["file"] not in skip:
            yield item["file"]


def check_field_write(f, md_ref, shape_ref, values_ref, use_writer=False, **kwargs):
    with temp_file() as tmp:
        if use_writer:
            fp = new_grib_output(tmp, template=f, **kwargs)
            fp.write(f.values)
            fp.close()
        else:
            f.save(tmp, **kwargs)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        assert len(r_tmp) == 1
        f = r_tmp[0]
        md = {k: f.metadata(k) for k in md_ref}
        md_ref_tmp = dict(**md_ref)

        # for single point data, the original bitsPerValue is 12. In the cloned handle it becomes 0.
        # For some reason when we try to set it again in the clone it results in 0!
        if len(values_ref) == 1:
            md_ref_tmp["bitsPerValue"] = 0

        assert md == md_ref_tmp

        assert f.shape == shape_ref
        assert len(f.values) == len(values_ref)
        if md["gridType"] != "sh":
            assert np.allclose(f.values, values_ref, rtol=1e-1)


# @pytest.mark.skipif(True, reason="headers_only_clone has to be fixed")
def test_grib_metadata_override_headers_only_true_core():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    ref_size = ds[0].metadata("totalLength")

    md1 = ds[0].metadata().override(headers_only_clone=True)
    assert isinstance(md1, WrappedMetadata)
    assert md1._handle is not None
    assert md1._handle != ds[0]._handle
    assert md1["totalLength"] - ref_size < -10
    assert md1["bitsPerValue"] == 16
    assert md1["shortName"] == "2t"
    assert md1["typeOfLevel"] == "surface"

    md2 = md1._hide_internal_keys()
    assert isinstance(md2, RestrictedGribMetadata)
    assert md2._handle is not None
    assert md2._handle != ds[0]._handle
    assert md2._handle == md1._handle
    assert md2.extra == {"bitsPerValue": 16}
    assert md2["bitsPerValue"] == 16
    assert md2["shortName"] == "2t"
    assert md2["typeOfLevel"] == "surface"

    with pytest.raises(KeyError):
        md2["average"]

    md3 = md2.override(headers_only_clone=True, shortName="2d")
    assert isinstance(md3, RestrictedGribMetadata)
    assert md3._handle is not None
    assert md3._handle != ds[0]._handle
    assert md3._handle != md1._handle
    assert md3._handle != md2._handle
    assert md3["totalLength"] - ref_size < -10
    assert md3.extra == {"bitsPerValue": 16}
    assert md3["bitsPerValue"] == 16
    assert md3["shortName"] == "2d"
    assert md3["typeOfLevel"] == "surface"

    md4 = md3._hide_internal_keys()
    assert md4 is md3

    md5 = md3.override(headers_only_clone=True, bitsPerValue=8)
    assert isinstance(md5, RestrictedGribMetadata)
    assert md5._handle is not None
    assert md5._handle != ds[0]._handle
    assert md5._handle != md1._handle
    assert md5._handle != md3._handle
    assert md5["totalLength"] - ref_size < -10
    assert md5.extra == {"bitsPerValue": 8}
    assert md5["bitsPerValue"] == 8
    assert md5["shortName"] == "2d"
    assert md5["typeOfLevel"] == "surface"


def test_grib_metadata_override_headers_only_false_core():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    ref_size = ds[0].metadata("totalLength")

    md1 = ds[0].metadata().override(headers_only_clone=False)
    assert isinstance(md1, WrappedMetadata)
    assert isinstance(md1.metadata, StandAloneGribMetadata)
    assert md1._handle is not None
    assert md1._handle != ds[0]._handle
    assert np.isclose(md1["totalLength"], ref_size)

    md2 = md1._hide_internal_keys()
    assert isinstance(md2, RestrictedGribMetadata)
    assert isinstance(md2.metadata, StandAloneGribMetadata)
    assert md2._handle is not None
    assert md2._handle != ds[0]._handle
    assert md2._handle == md1._handle

    with pytest.raises(KeyError):
        md2["average"]


@pytest.mark.cache
@pytest.mark.parametrize(
    "file",
    # grid_list(files=["regular_ll_single_point.grib1"]),
    grid_list(),
)
def test_grib_metadata_headers_only_clone_true_grids(file):
    ds0 = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "grid", file))

    keys = ["bitsPerValue", "level", "shortName", "gridType", "packingType", "date"]
    md_ref = {k: ds0[0].metadata(k) for k in keys}
    shape_ref = ds0[0].shape
    vals_ref = ds0[0].values

    # array field
    ds = ds0.to_fieldlist()
    md = {k: ds[0].metadata(k) for k in keys}
    assert isinstance(ds[0].metadata(), RestrictedGribMetadata)
    assert md == md_ref
    assert ds[0].shape == shape_ref

    # create new field with modified metadata
    f = ds[0].copy(metadata=ds[0].metadata().override(date=19810102))
    assert isinstance(f.metadata(), RestrictedGribMetadata)
    md_ref_1 = dict(**md_ref)
    md_ref_1["date"] = 19810102
    md = {k: f.metadata(k) for k in keys}
    assert md == md_ref_1

    # save to disk, should use the original bitsPerValue

    # for single point data, the original bitsPerValue is 12. In the cloned handle it becomes 0.
    # For some reason when we try to set it again to 12 on the clone it results in 0!
    check_field_write(f, md_ref_1, shape_ref, vals_ref)

    # save to disk, with the given bitsPerValue
    md_ref_tmp = dict(**md_ref_1)
    md_ref_tmp["bitsPerValue"] = 24
    check_field_write(f, md_ref_tmp, shape_ref, vals_ref, bitsPerValue=24)

    # create new field with modified bitsPerValue
    f = f.copy(metadata=f.metadata().override(bitsPerValue=8))
    assert isinstance(f.metadata(), RestrictedGribMetadata)
    md_ref_1["bitsPerValue"] = 8
    md = {k: f.metadata(k) for k in keys}
    assert md == md_ref_1

    # save to disk, with the given bitsPerValue
    check_field_write(f, md_ref_1, shape_ref, vals_ref)

    # save to disk, with the given bitsPerValue
    check_field_write(f, md_ref_1, shape_ref, vals_ref, use_writer=True)


@pytest.mark.cache
@pytest.mark.parametrize(
    "file",
    # grid_list(files=["reduced_gg_N32.grib1"]),
    grid_list(),
)
def test_grib_metadata_headers_only_clone_false_grids(file):
    ds0 = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "grid", file))

    keys = ["bitsPerValue", "level", "shortName", "gridType", "packingType", "date"]
    md_ref = {k: ds0[0].metadata(k) for k in keys}
    shape_ref = ds0[0].shape
    vals_ref = ds0[0].values

    # array field
    ds = ds0.from_fields([ds0[0].copy(metadata=ds0[0].metadata().override(headers_only_clone=False))])
    md = {k: ds[0].metadata(k) for k in keys}
    assert isinstance(ds[0].metadata(), RestrictedGribMetadata)
    assert md == md_ref
    assert ds[0].shape == shape_ref

    # create new field with modified metadata
    f = ds[0].copy(metadata=ds[0].metadata().override(headers_only_clone=False, date=19810102))
    assert isinstance(f.metadata(), RestrictedGribMetadata)
    md_ref_1 = dict(**md_ref)
    md_ref_1["date"] = 19810102
    md = {k: f.metadata(k) for k in keys}
    assert md == md_ref_1

    # save to disk, should use the original bitsPerValue

    # for single point data, the original bitsPerValue is 12. In the cloned handle it becomes 0.
    # For some reason when we try to set it again to 12 on the clone it results in 0!
    check_field_write(f, md_ref_1, shape_ref, vals_ref)

    # save to disk, with the given bitsPerValue
    md_ref_tmp = dict(**md_ref_1)
    md_ref_tmp["bitsPerValue"] = 24
    check_field_write(f, md_ref_tmp, shape_ref, vals_ref, bitsPerValue=24)

    # create new field with modified bitsPerValue
    f = f.copy(metadata=f.metadata().override(bitsPerValue=8, headers_only_clone=False))
    assert isinstance(f.metadata(), RestrictedGribMetadata)
    md_ref_1["bitsPerValue"] = 8
    md = {k: f.metadata(k) for k in keys}
    assert md == md_ref_1

    # save to disk, with the given bitsPerValue
    check_field_write(f, md_ref_1, shape_ref, vals_ref)

    # save to disk, with the given bitsPerValue
    check_field_write(f, md_ref_1, shape_ref, vals_ref, use_writer=True)


def test_grib_headers_only_clone_standalone_metadata():
    ds = from_source("file", earthkit_examples_file("test.grib"))

    md_ref = {
        "param": "2t",
        "date": 20200513,
        "time": 1200,
        "step": 0,
        "level": 0,
        "gridType": "regular_ll",
        "type": "an",
    }

    md0 = ds[0].metadata().override()
    md1 = StandAloneGribMetadata(md0._handle)
    for k, v in md_ref.items():
        assert md1[k] == v

    # the handle does not contain bitsPerValue
    assert md1["bitsPerValue"] == 0

    md0 = ds[0].metadata().override(bitsPerValue=8)
    md1 = StandAloneGribMetadata(md0._handle)
    for k, v in md_ref.items():
        assert md1[k] == v

    # the handle does not contain bitsPerValue
    assert md1["bitsPerValue"] == 0



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import re

from . import SimpleTarget

LOG = logging.getLogger(__name__)


class FilePatternTarget(SimpleTarget):
    """
    File target with a pattern for the output file names.

    Parameters:
    -----------
    path: str
        The file path to write to. The output file name defines a pattern containing metadata keys in the
        format of ``{key}``. Each data item (e.g. a field) will be written into a file
        with a name created by substituting the relevant metadata values in the
        filename pattern.
    append: bool
        If True, the files are opened in append mode. Only used if file is a path.
    **kwargs:
        Additional keyword arguments passed to the parent class
    """

    def __init__(self, path, *, append=False, **kwargs):
        super().__init__(**kwargs)

        self._files = {}
        self.filename = path
        self.append = append
        self.ext = None

        if not self.filename:
            raise ValueError("Please provide an output filename")

        _, self.ext = os.path.splitext(self.filename)

        self.split_output = re.findall(r"\{(.*?)\}", self.filename)

        # change format pattern to use positional arguments. This allows using keys
        # containing a dot, e.g. mars.param
        for i, s in enumerate(self.split_output):
            if ":" in s:
                pos_key = f"{i}:" + s.split(":")[1]
            else:
                pos_key = f"{i}"

            self.filename = self.filename.replace(f"{s}", pos_key)

    def close(self):
        """Close the target and closing all the files.

        The target will not be able to write anymore.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        self._mark_closed()
        for f in self._files.values():
            f.close()

    def flush(self):
        """Flush all the files.

        -------
        ValueError: If the target is already closed.
        """
        for f in self._files.values():
            f.flush()

    def _f(self, data):
        self._raise_if_closed()

        keys = [data.metadata(k.split(":")[0]) for k in self.split_output]
        path = self.filename.format(*keys)

        if path not in self._files:
            flag = "wb" if not self.append else "ab"
            self._files[path] = open(path, flag)

        return self._files[path], path

    def _write(self, data, **kwargs):
        r = self._encode(data, suffix=self.ext, **kwargs)
        if hasattr(r, "__iter__"):
            for d in r:
                f, _ = self._f(d)
                d.to_file(f)
        else:
            f, _ = self._f(r)
            r.to_file(f)


target = FilePatternTarget



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
from abc import ABCMeta
from abc import abstractmethod


class MetadataAccessor:
    r"""Helper class to access metadata using dedicated methods on a metadata object.

    Parameters
    ----------
    conf: dict
        Dictionary mapping method names to metadata keys that should be accessed with the given method.m
        The metadata keys can be a string or a list/tuple of strings.
    aliases: list of iterables, optional
        Groups of metadata keys that should be accessed with the same method.
    """

    def __init__(self, conf, aliases=None):
        self.keys = set()
        self.methods = dict()
        aliases = aliases or dict()
        for k, v in conf.items():
            if isinstance(v, str):
                self.keys.add(v)
            else:
                self.keys.update(v)

            if isinstance(v, str):
                v = [v]

            for kv in v:
                self.methods[kv] = k

            for gr in aliases:
                if any(gr_k in v for gr_k in gr):
                    for gr_k in gr:
                        if gr_k not in self.keys:
                            self.methods[gr_k] = k
                            self.keys.add(gr_k)

    def __contains__(self, key):
        return key in self.keys

    def get(self, md, key, default=None, *, astype=None, raise_on_missing=False):
        """Return the value for ``key``.

        Parameters
        ----------
        md: Metadata
            Metadata object featuring the method to access the value of ``key``.
        key: str
            Metadata key
        """
        try:
            v = getattr(md, self.methods[key])()
            # TODO: add automatic conversion support
            if self.methods[key].endswith("_datetime"):
                if v is not None:
                    return v.isoformat()
            return v
        except Exception as e:
            if not raise_on_missing:
                return default
            else:
                raise KeyError(f"{key}, reason={e}")

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped(cls, key, *args, **kwargs):
            if key in self:
                return self.get(cls, key, *args, **kwargs)
            return func(cls, key, *args, **kwargs)

        return wrapped


class MetadataCacheHandler:
    @staticmethod
    def make(cache=None):
        if cache is True:
            return dict()
        elif cache is not False and cache is not None:
            return cache

    @staticmethod
    def clone_empty(cache):
        if cache is not None:
            return cache.__class__()

    @staticmethod
    def serialise(cache):
        if cache is not None:
            return cache.__class__

    @staticmethod
    def deserialise(state):
        cache = state
        if state is not None:
            return cache()

    @staticmethod
    def cache_get(func):
        @functools.wraps(func)
        def wrapped(self, key, default=None, *, astype=None, raise_on_missing=False):
            if self._cache is not None:
                cache_id = (key, default, astype, raise_on_missing)
                if cache_id in self._cache:
                    return self._cache[cache_id]

                v = func(self, key, default=default, astype=astype, raise_on_missing=raise_on_missing)
                self._cache[cache_id] = v
                return v
            else:
                return func(self, key, default=default, astype=astype, raise_on_missing=raise_on_missing)

        return wrapped


class Metadata(metaclass=ABCMeta):
    r"""Base class to represent metadata.

    Metadata is a dict-like immutable object. Modification is possible
    with :obj:`override`, which always creates a new object.

    Implemented in subclasses: :obj:`RawMetadata`, :obj:`GribMetadata`.

    Parameters
    ----------
    cache: bool, dict-like, optional
        Enable caching of all the calls to :meth:`get`. Default is False. The cache
        is attached to the instance.

    Examples
    --------
    - :ref:`/examples/metadata.ipynb`

    """

    @abstractmethod
    def __iter__(self):
        """Return an iterator over the metadata keys."""
        pass

    @abstractmethod
    def __len__(self):
        r"""Return the number of metadata entries."""
        pass

    def __getitem__(self, key):
        r"""Return the value for ``key``.

        Raises
        ------
        KeyError
            When ``key`` is not available.
        """
        return self.get(key, raise_on_missing=True)

    @abstractmethod
    def __contains__(self, key):
        r"""Check if ``key`` is available.

        Returns
        -------
        bool
        """
        pass

    @abstractmethod
    def keys(self):
        r"""Return the metadata keys.

        Returns
        -------
        Iterable of str

        """
        pass

    @abstractmethod
    def items(self):
        r"""Return the metadata items.

        Returns
        -------
        Iterable of :obj:`(key,value)` pairs

        """
        pass

    @abstractmethod
    def get(self, key, default=None, *, astype=None, raise_on_missing=False):
        r"""Return the value for ``key``.

        Parameters
        ----------
        key: str
            Metadata key
        default: value
            Specify the default value for ``key``. Returned when ``key``
            is not found or its value is a missing value and raise_on_missing is ``False``.
        astype: type as str, int or float
            Return/access type for ``key``. When it is supported ``astype`` is passed to the
            underlying metadata accessor as an option. Otherwise the value is
            cast to ``astype`` after it is taken from the accessor.
        raise_on_missing: bool
            When it is True raises an exception if ``key`` is not found or
            it has a missing value.

        Returns
        -------
        value
            Returns the ``key`` value. Returns ``default`` if ``key`` is not found
            or it has a missing value and ``raise_on_missing`` is False.

        Raises
        ------
        KeyError
            If ``raise_on_missing`` is True and ``key`` is not found or it has
            a missing value.

        """
        pass

    @abstractmethod
    def override(self, *args, **kwargs):
        r"""Change the metadata values and return a new object.

        :obj:`override` accepts another :obj:`Metadata` or a dict or
        an iterable of key/value pairs (as tuples or other iterables of length two).
        If keyword arguments are specified, the metadata is then updated with those
        key/value pairs.

        Examples
        --------
        >>> other = RawMetadata({"key1": 1, "key2": 2})
        >>> m1 = m.override(other)
        >>> m1 = m.override({"key1": 1, "key2": 2})
        >>> m1 = m.override([("key1", 1), ("key2", 2)])
        >>> m1 = m.override(key1=1, key2=2)
        """
        pass

    @abstractmethod
    def namespaces(self):
        r"""Return the available namespaces.

        Returns
        -------
        list of str
        """
        pass

    @abstractmethod
    def as_namespace(self, namespace=None):
        r"""Return all the keys/values from a namespace.

        Parameters
        ----------
        namespace: str, None
            The namespace. When `namespace` is None or an empty :obj:`str` all the available
            keys/values are returned.

        Returns
        -------
        dict
            All the keys/values from the `namespace`.

        """
        pass

    @abstractmethod
    def dump(self, **kwargs):
        r"""Generate a dump from the metadata content."""
        pass

    @abstractmethod
    def datetime(self):
        r"""Return the date and time of the field.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds[4].datetime()
        {'base_time': datetime.datetime(2020, 12, 21, 12, 0),
        'valid_time': datetime.datetime(2020, 12, 21, 18, 0)}

        """
        pass

    @abstractmethod
    def base_datetime(self):
        pass

    @abstractmethod
    def valid_datetime(self):
        pass

    @property
    @abstractmethod
    def geography(self):
        r""":obj:`Geography`: Get geography description.

        If it is not available None is returned.
        """
        pass

    @abstractmethod
    def ls_keys(self):
        r"""Return the keys to be used with the :meth:`ls` method."""
        pass

    @abstractmethod
    def describe_keys(self):
        r"""Return the keys to be used with the :meth:`describe` method."""
        pass

    @abstractmethod
    def index_keys(self):
        r"""Return the keys to be used with the :meth:`indices` method."""
        pass

    @abstractmethod
    def data_format(self):
        r"""Return the underlying data format.

        Returns
        -------
        str

        """
        pass

    @abstractmethod
    def _hide_internal_keys(self):
        """If the metadata object has internal keys, hide them.

        Returns
        -------
        WrappedMetadata, Metadata
            If the metadata object has internal keys, return a new wrapped object with the
            internal keys hidden. Otherwise return the metadata object itself.
        """
        pass


class WrappedMetadata:
    r"""Wrapper around a Metadata object to add extra metadata entries and hide some keys.

    Parameters
    ----------
    metadata: Metadata
        The metadata object to wrap.
    extra: dict-like, optional
        Additional metadata entries, which are not part of ``metadata``.
        Methods like `:obj:get` first search for entries in `extra` before
        querying ``metadata``. The ``extra`` entries can be callables that
        take the following arguments: owner, key, metadata (where metadata
        is the wrapped metadata object).
    hidden: list of str, None, optional
        Metadata keys to hide from ``metadata``.
    hidden_namespaces: list of str, None, optional
        Namespaces to hide from ``metadata``.
    enforced_namespaces: list of str, None, optional
        Keys in these namespaces are cannot be hidden.
    owner: object, optional
        The owner.
    merge: bool, optional
        Used when True and ``metadata`` is a :class:`WrappedMetadata`. In this case
        merge ``extra``, ``hidden``, ``hidden_namespaces``, and ``enforced_namespaces``
        from the original metadata and also replaces it with the object the original
        metadata wrapped.

    Raises
    ------
    ValueError: If a key in ``hidden`` is also in ``extra``.
    """

    def __init__(
        self,
        metadata,
        extra=None,
        hidden=None,
        hidden_namespaces=None,
        enforced_namespaces=None,
        owner=None,
        merge=True,
    ):
        self.metadata = metadata
        self.extra = extra if extra is not None else dict()
        self.hidden = hidden if hidden is not None else []
        self.hidden_namespaces = hidden_namespaces
        self.enforced_namespaces = enforced_namespaces
        self.owner = owner

        for k in self.hidden:
            if k in self.extra:
                raise ValueError(f"Hidden key {k} is also in extra")

        if merge and isinstance(metadata, WrappedMetadata):
            self.metadata = metadata.metadata
            self._update(metadata)

    def __len__(self):
        r"""Return the number of metadata entries."""
        return len(self.keys())

    def __contains__(self, key):
        r"""Check if ``key`` is available.

        Returns
        -------
        bool
        """
        if self._is_hidden(key):
            return False
        if key in self.extra:
            return True
        return key in self.metadata

    def _is_hidden(self, key):
        name = key
        if "." in key:
            ns, _, name = key.partition(".")
            if name == "":
                name = key
                ns = ""
            if ns and self.enforced_namespaces and ns in self.enforced_namespaces:
                return False

        return name in self.hidden

    def keys(self):
        r"""Return the metadata keys.

        Returns
        -------
        Iterable of str

        """
        r = list(self.metadata.keys())
        if self.extra:
            r += [x for x in self.extra if x not in r]
        if self.hidden:
            r = [x for x in r if x not in self.hidden]
        return r

    def items(self):
        r"""Return the metadata items.

        Returns
        -------
        Iterable of :obj:`(key,value)` pairs

        """
        for k in self.keys():
            if self._is_hidden(k):
                continue
            if k in self.extra:
                yield k, self.extra[k]
            else:
                yield k, self.metadata.get(k)

    def get(self, key, default=None, *, astype=None, raise_on_missing=False, **kwargs):
        if self._is_hidden(key):
            if raise_on_missing:
                raise KeyError(key)
            return default

        if key in self.extra:
            v = self._extra_value(key)
            if astype is not None and v is not None:
                try:
                    return astype(v)
                except Exception:
                    return None
            return v

        return self.metadata.get(
            key, default=default, astype=astype, raise_on_missing=raise_on_missing, **kwargs
        )

    def _extra_value(self, key):
        v = self.extra[key]
        if callable(v):
            v = v(self.owner, key, self.metadata)
        return v

    def namespaces(self):
        if self.hidden_namespaces:
            return [x for x in self.metadata.namespaces() if x not in self.hidden_namespaces]
        else:
            return self.metadata.namespaces()

    def as_namespace(self, namespace):
        if self.hidden_namespaces and namespace in self.hidden_namespaces:
            return {}

        r = dict()
        if namespace is None:
            r = dict(self.items())
            for k, _ in self.extra.items():
                if k in r:
                    r[k] = self._extra_value(k)
        else:
            r = self.metadata.as_namespace(namespace)
            for k in list(r.keys()):
                if k in self.hidden:
                    del r[k]
            # TODO: add filtering based on extra

        return r

    def dump(self, namespace=all, **kwargs):
        if namespace is all:
            namespace = self.namespaces()
        return self.metadata.dump(namespace=namespace, **kwargs)

    def override(self, *args, **kwargs):
        md = self.metadata.override(*args, **kwargs)
        if self.metadata is md:
            return self
        return self._clone(md)

    def _hide_internal_keys(self):
        if self.hidden:
            return self

        md = self.metadata._hide_internal_keys()
        if self.metadata is md:
            return self
        elif isinstance(md, WrappedMetadata) and md.metadata is self.metadata:
            md._update(self)
            return md
        else:
            return self._clone(md)

    @staticmethod
    def merge_list(v1, v2):
        if v1 and v2:
            if v2 is None:
                v2 = list(v1)
                return v2
            r = [x for x in v1 if x not in v2]
            if r:
                v2 = list(v2)
                v2.extend(r)

    def _update(self, other):
        assert isinstance(other, WrappedMetadata)
        v = dict(**other.extra)
        # self.extra.update(other.extra)
        v.update(self.extra)
        self.extra = v
        self.merge_list(other.hidden, self.hidden)
        self.merge_list(other.hidden_namespaces, self.hidden_namespaces)
        self.merge_list(other.enforced_namespaces, self.enforced_namespaces)

    def _clone(self, metadata):
        r = self.__class__(metadata)
        r._update(self)
        return r

    def __getitem__(self, key):
        return self.get(key, raise_on_missing=True)

    def __getattr__(self, name):
        return getattr(self.metadata, name)

    def __getstate__(self) -> dict:
        ret = {}
        ret["metadata"] = self.metadata
        ret["extra"] = self.extra
        ret["hidden"] = self.hidden
        ret["hidden_namespaces"] = self.hidden_namespaces
        ret["enforced_namespaces"] = self.enforced_namespaces
        return ret

    def __setstate__(self, state: dict):
        self.metadata = state.pop("metadata")
        self.extra = state.pop("extra")
        self.hidden = state.pop("hidden")
        self.hidden_namespaces = state.pop("hidden_namespaces")
        self.enforced_namespaces = state.pop("enforced_namespaces")


class RawMetadata(Metadata):
    r"""Metadata implementation based on key/value pairs.

    >>> from earthkit.data.core.metadata import RawMetadata
    >>> md = RawMetadata({"shortName": "2t", "perturbationNumber": 5})
    >>> md = RawMetadata([("shortName", "2t"), ("perturbationNumber", 5)])
    >>> md = RawMetadata(shortName="2t", perturbationNumber=5)
    >>> md
    RawMetadata({'shortName': '2t', 'perturbationNumber': 5})
    >>> md2 = RawMetadata(md)
    >>> md2
    RawMetadata({'shortName': '2t', 'perturbationNumber': 5})

    Value access:

    >>> "shortName" in md
    True
    >>> md["shortName"]
    '2t'
    >>> md.get("shortName")
    '2t'
    >>> "step" in md
    False
    >>> md.get("step")
    >>> md.get("step", 0)
    0

    Examples
    --------
    - :ref:`/examples/metadata.ipynb`

    """

    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)
        super().__init__()

    def __len__(self):
        return len(self._d)

    def __contains__(self, key):
        return key in self._d

    def __getitem__(self, key):
        return self.get(key, raise_on_missing=True)

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        return self._d.keys()

    def items(self):
        return self._d.items()

    def get(self, key, default=None, *, astype=None, raise_on_missing=False):
        if not raise_on_missing:
            v = self._d.get(key, default)
        else:
            v = self._d[key]

        if astype is not None:
            try:
                return astype(v)
            except Exception:
                return None
        return v

    def override(self, *args, **kwargs):
        d = dict(**self._d)
        d.update(*args, **kwargs)
        return RawMetadata(d)

    def namespaces(self):
        return []

    def as_namespace(self, namespace):
        return {}

    def datetime(self):
        return None

    def base_datetime(self):
        return None

    def valid_datetime(self):
        return None

    @property
    def geography(self):
        return None

    def dump(self, **kwargs):
        return None

    def ls_keys(self):
        return []

    def describe_keys(self):
        return []

    def index_keys(self):
        return []

    def data_format(self):
        return None

    def _hide_internal_keys(self):
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}({self._d.__repr__()})"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools
import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import all_params  # noqa: E402


def test_forcings_source_1():
    sample = from_source("file", earthkit_examples_file("test.grib"))

    start = sample[0].datetime()["valid_time"]
    first_step = 6
    last_step = 240
    step_increment = 6
    dates = []
    for step in range(first_step, last_step + step_increment, step_increment):
        dates.append(start + datetime.timedelta(hours=step))

    params = all_params

    ds = from_source(
        "forcings",
        sample,
        date=dates,
        param=all_params,
    )

    num = len(params) * len(dates)
    assert len(ds) == num

    ref = [(d, p) for d, p in itertools.product(dates, params)]
    assert len(ds) == len(ref)
    for f, r in zip(ds, ref):
        assert f.metadata("valid_datetime") == r[0].isoformat()
        assert f.metadata("param") == r[1]


def test_forcings_2():
    sample = from_source("file", earthkit_examples_file("test.grib"))

    start = sample[0].datetime()["valid_time"]
    start = datetime.datetime(start.year, start.month, start.day)
    first_step = 1
    last_step = 10
    step_increment = 1
    dates = []
    for step in range(first_step, last_step + step_increment, step_increment):
        dates.append(start + datetime.timedelta(days=step))

    params = all_params

    ntimes = 4
    ds = from_source(
        "forcings",
        sample,
        date=dates,
        time=f"0/to/18/by/{24//ntimes}",
        param=params,
    )

    num = len(params) * len(dates) * 4
    assert len(ds) == num

    ref = [(d, p) for d, p in itertools.product(ds.dates, params)]
    assert len(ds) == len(ref)
    for f, r in zip(ds, ref):
        assert f.metadata("valid_datetime") == r[0].isoformat()
        assert f.metadata("param") == r[1]


def test_forcings_3():
    sample = from_source("file", earthkit_test_data_file("t_time_series.grib"))

    dates = [
        datetime.datetime(2020, 12, 21, 12, 0),
        datetime.datetime(2020, 12, 21, 15, 0),
        datetime.datetime(2020, 12, 21, 18, 0),
        datetime.datetime(2020, 12, 21, 21, 0),
        datetime.datetime(2020, 12, 23, 12, 0),
    ]

    params = all_params

    ds = from_source(
        "forcings",
        sample,
        param=params,
    )

    num = len(dates) * len(params)
    assert len(ds) == num

    ref = [(d, p) for d, p in itertools.product(dates, params)]
    assert len(ds) == len(ref)
    for f, r in zip(ds, ref):
        assert f.metadata("valid_datetime") == r[0].isoformat()
        assert f.metadata("param") == r[1]


@pytest.mark.parametrize("lat_key,lon_key", [("latitudes", "longitudes"), ("latitude", "longitude")])
@pytest.mark.parametrize(
    "filename", ["t_time_series.grib", "rgg_small_subarea_cellarea_ref.grib", "mercator.grib"]
)
def test_forcings_from_lat_lon_core(lat_key, lon_key, filename):
    sample = from_source("file", earthkit_test_data_file(filename))

    dates = [
        datetime.datetime(2020, 12, 21, 12, 0),
        datetime.datetime(2020, 12, 21, 15, 0),
        datetime.datetime(2020, 12, 21, 18, 0),
        datetime.datetime(2020, 12, 21, 21, 0),
        datetime.datetime(2020, 12, 23, 12, 0),
    ]

    params = all_params

    ll = sample[0].to_latlon()  # flatten=True is important here
    lats = ll["lat"]
    lons = ll["lon"]

    d = {}
    d[lat_key] = lats
    d[lon_key] = lons

    ds = from_source("forcings", **d, date=dates, param=params)

    num = len(dates) * len(params)
    assert len(ds) == num

    ref = [(d, p) for d, p in itertools.product(dates, params)]
    assert len(ds) == len(ref)
    for f, r in zip(ds, ref):
        assert f.metadata("valid_datetime") == r[0].isoformat()
        assert f.metadata("param") == r[1]
        assert f.to_numpy().shape == lats.shape


def test_forcings_from_lat_lon_bad():
    sample = from_source("file", earthkit_test_data_file("t_time_series.grib"))

    params = all_params

    ll = sample[0].to_latlon()
    lats = ll["lat"]
    lons = ll["lon"]

    with pytest.raises(ValueError):
        from_source(
            "forcings",
            latitudes=lats,
            param=params,
        )
    with pytest.raises(ValueError):
        from_source(
            "forcings",
            latitude=lats,
            param=params,
        )
    with pytest.raises(ValueError):
        from_source(
            "forcings",
            longitudes=lons,
            param=params,
        )
    with pytest.raises(ValueError):
        from_source(
            "forcings",
            longitude=lons,
            param=params,
        )


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from earthkit.data.translators import Translator


class PandasSeriesTranslator(Translator):
    """Translator class for pandas `Series`"""

    def __init__(self, data, *args, **kwargs):
        self.data = data.to_pandas(*args, **kwargs)

    def __call__(self):
        """Series requested, if DataFrame return the first column"""
        import pandas as pd

        if isinstance(self.data, pd.DataFrame):
            return self.data.iloc[:, 0]

        return self.data


class PandasDataFrameTranslator(PandasSeriesTranslator):
    """Translator class for pandas `DataFrame`"""

    def __call__(self):
        """Return DataFrame, if Series convert to DataFrame."""
        import pandas as pd

        if isinstance(self.data, pd.Series):
            return self.data.to_frame()

        return self.data


class GeoPandasDataFrameTranslator(PandasSeriesTranslator):
    """Translator class for geopandas `DataFrame`"""

    def __call__(self):
        """Return GeoDataFrame, if normal pandas convert to geopandas."""
        import geopandas as gpd
        import pandas as pd

        if isinstance(self.data, pd.DataFrame):
            return gpd.GeoDataFrame(self.data)

        return self.data


def translator(data, cls, *args, **kwargs):
    from earthkit.data.utils import is_module_loaded

    if not is_module_loaded("pandas"):
        return None

    import pandas as pd

    if is_module_loaded("geopandas"):
        import geopandas as gpd

        if cls in [gpd.geodataframe.GeoDataFrame]:
            return GeoPandasDataFrameTranslator(data, *args, **kwargs)

    if cls in [pd.DataFrame]:
        return PandasDataFrameTranslator(data, *args, **kwargs)
    if cls in [pd.Series]:
        return PandasSeriesTranslator(data, *args, **kwargs)

    return None



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import config
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file


def read_config_yaml(path=os.path.expanduser("~/.config/earthkit/data/config.yaml")):
    try:
        with open(path) as f:
            import yaml

            s = yaml.load(f, Loader=yaml.SafeLoader)
            if not isinstance(s, dict):
                s = {}
            return s
    except Exception:
        return {}


@pytest.mark.parametrize(
    "param,default_value,new_value",
    [
        ("url-download-timeout", 30, 5),
    ],
)
def test_configs_params_set_reset(param, default_value, new_value):
    ori_value = config.get(param)

    with config.temporary():
        config.reset()

        assert config.get(param) == default_value
        config.set(param, new_value)
        assert config.get(param) == new_value
        config.reset()
        assert config.get(param) == default_value

    assert config.get(param) == ori_value


def test_config_invalid():
    # invalid param
    with pytest.raises(KeyError):
        config.get("_invalid_")

    with pytest.raises(KeyError):
        config.set("_invalid_", 1)

    # invalid value
    with pytest.raises(ValueError):
        config.set("url-download-timeout", "A")


@pytest.mark.parametrize(
    "param,set_value,stored_value,raise_error",
    [
        ("number-of-download-threads", "A", "A", ValueError),
        ("url-download-timeout", 30, 30, None),
        ("url-download-timeout", "30", 30, None),
        ("url-download-timeout", "2m", 120, None),
        ("url-download-timeout", "10h", 36000, None),
        ("url-download-timeout", "7d", 7 * 24 * 3600, None),
        ("url-download-timeout", "1x", None, ValueError),
        ("url-download-timeout", "1M", 60, ValueError),
        ("reader-type-check-bytes", 8, 8, None),
        ("reader-type-check-bytes", 1, 1, ValueError),
        ("reader-type-check-bytes", 4097, 4097, ValueError),
    ],
)
def test_config_set_numbers(param, set_value, stored_value, raise_error):
    with config.temporary():
        if raise_error is None:
            config.set(param, set_value)
            assert config.get(param) == stored_value
        else:
            with pytest.raises(raise_error):
                config.set(param, set_value)


def test_config_set_cache_numbers():
    with temp_directory() as tmpdir:
        with config.temporary({"cache-policy": "user", "user-cache-directory": tmpdir}):
            data = [
                ("maximum-cache-size", "1", 1, None),
                ("maximum-cache-size", "1k", 1024, None),
                ("maximum-cache-size", "1kb", 1024, None),
                ("maximum-cache-size", "1k", 1024, None),
                ("maximum-cache-size", "1kb", 1024, None),
                ("maximum-cache-size", "1K", 1024, None),
                ("maximum-cache-size", "1M", 1024 * 1024, None),
                ("maximum-cache-size", "1G", 1024 * 1024 * 1024, None),
                ("maximum-cache-size", "1T", 1024 * 1024 * 1024 * 1024, None),
                ("maximum-cache-size", "1P", 1024 * 1024 * 1024 * 1024 * 1024, None),
                ("maximum-cache-size", None, None, None),
                ("maximum-cache-size", "-1", None, ValueError),
                ("maximum-cache-disk-usage", "2%", 2, None),
                ("maximum-cache-disk-usage", None, None, None),
                ("maximum-cache-disk-usage", "-2%", None, ValueError),
                ("maximum-cache-disk-usage", "102%", 102, None),
                ("maximum-cache-disk-usage", "0%", 0, None),
            ]

            for param, set_value, stored_value, raise_error in data:
                if raise_error is None:
                    config.set(param, set_value)
                    if stored_value is not None:
                        assert config.get(param) == stored_value
                    else:
                        assert config.get(param) is None
                else:
                    with pytest.raises(raise_error):
                        config.set(param, set_value)


def test_config_set_multi():
    with config.temporary():
        config.set("url-download-timeout", 7)
        assert config.get("url-download-timeout") == 7

        config.set({"url-download-timeout": 2, "check-out-of-date-urls": False})
        assert config.get("url-download-timeout") == 2
        assert not config.get("check-out-of-date-urls")

        config.set(url_download_timeout=11, check_out_of_date_urls=False)
        assert config.get("url-download-timeout") == 11
        assert not config.get("check-out-of-date-urls")

        with pytest.raises(KeyError):
            config.set({"url-download-timeout": 2, "-invalid-": 21})

        with pytest.raises(KeyError):
            config.set(url_download_timeout=3, __invalid__=11)


def test_config_temporary_single():
    with config.temporary("url-download-timeout", 7):
        assert config.get("url-download-timeout") == 7

    with config.temporary({"url-download-timeout": 7}):
        assert config.get("url-download-timeout") == 7

    with config.temporary(url_download_timeout=7):
        assert config.get("url-download-timeout") == 7


def test_config_temporary_multi():
    with config.temporary({"url-download-timeout": 2, "check-out-of-date-urls": False}):
        assert config.get("url-download-timeout") == 2
        assert not config.get("check-out-of-date-urls")

    with config.temporary(url_download_timeout=3, check_out_of_date_urls=False):
        assert config.get("url-download-timeout") == 3
        assert not config.get("check-out-of-date-urls")


def test_config_temporary_nested():
    with config.temporary("url-download-timeout", 7):
        assert config.get("url-download-timeout") == 7
        with config.temporary("url-download-timeout", 10):
            assert config.get("url-download-timeout") == 10
        assert config.get("url-download-timeout") == 7


def test_config_temporary_autosave_1():
    with temp_file() as config_file:
        with config.temporary(config_yaml=config_file):
            # now config should contain the default values
            # we ensure that the configs are saved into the file
            config.save_as(config_file)

            key = "url-download-timeout"

            v_ori = config.autosave
            config.autosave = False

            # when a key has a default value, it is not saved into the config file
            s = read_config_yaml(config_file)
            assert key not in s

            v = config.get(key)
            config.set(key, v + 10)
            assert config.get(key) == v + 10

            # the config file should be the same
            s = read_config_yaml(config_file)
            assert key not in s

            config.autosave = v_ori


def test_config_temporary_autosave_2():
    with temp_file() as config_file:
        with config.temporary(config_yaml=config_file):
            # now config should contain the default values
            # we ensure that the config is saved into the file
            config.save_as(config_file)

            key = "url-download-timeout"

            v_ori = config.autosave
            config.autosave = True

            # when a key has a default value, it is not saved into the config file
            s = read_config_yaml(config_file)
            assert key not in s

            v = config.get(key)
            v_new = v + 10
            config.set(key, v_new)
            assert config.get(key) == v_new

            # the file changed
            s = read_config_yaml(config_file)
            assert s[key] == v_new

            config.autosave = False
            config.set(key, v)
            assert config.get(key) == v
            s = read_config_yaml(config_file)
            assert s[key] == v_new

            config.autosave = v_ori


@pytest.mark.parametrize(
    "value,error", [("10000", None), (10000, None), ("1b", ValueError), ("A", ValueError)]
)
def test_config_env(monkeypatch, value, error):
    env_key = "EARTHKIT_DATA_URL_DOWNLOAD_TIMEOUT"
    monkeypatch.setenv(env_key, value)

    # v_ori = config.autosave
    # config.autosave = True

    if error is None:
        v = config.get("url-download-timeout")
        assert v == 10000
    else:
        with pytest.raises(error):
            config.get("url-download-timeout")

    # config.autosave = v_ori


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.core.order import build_remapping
from earthkit.data.utils.xarray.fieldlist import CollectorJoiner

data = {"type": "cf", "number": 0, "name": "t2m"}


def fn(key, **kwargs):
    return data[key]


@pytest.mark.parametrize(
    "remapping,expected_values",
    [
        ({}, {"type": "cf", "number": 0, "name": "t2m"}),
        (None, {"type": "cf", "number": 0, "name": "t2m"}),
        ({"type": "pf"}, {"type": "pf", "number": 0, "name": "t2m"}),
        ({"type": "{type}{number}"}, {"type": "cf0", "name": "t2m"}),
        ({"type": "{type}_{number}"}, {"type": "cf_0", "name": "t2m"}),
        ({"type": lambda: 12}, {"type": 12, "name": "t2m"}),
    ],
)
def test_remapping_dict(remapping, expected_values):
    r = build_remapping(remapping)
    _fn = r(fn)

    for k, v in expected_values.items():
        assert _fn(k) == v


def test_remapping_repeated():
    remapping = {"type": "{type}{number}"}
    r = build_remapping(remapping)
    r = build_remapping(r)
    _fn = r(fn)

    assert _fn("type") == "cf0"


@pytest.mark.parametrize(
    "remapping,patch,expected_values",
    [
        (
            {},
            {"type": {"cf": "pf"}, "number": 12, "name": None},
            {"type": "pf", "number": 12, "name": None},
        ),
        (
            {"type": "{type}{number}"},
            {"type": {"cf0": "x"}},
            {"type": "x", "number": 0, "name": "t2m"},
        ),
        ({}, {"type": lambda x: "x"}, {"type": "x", "number": 0, "name": "t2m"}),
        ({}, {"type": True}, {"type": True, "number": 0, "name": "t2m"}),
        (None, {"type": True}, {"type": True, "number": 0, "name": "t2m"}),
    ],
)
def test_remapping_patch(remapping, patch, expected_values):
    r = build_remapping(remapping, patch)
    _fn = r(fn)

    for k, v in expected_values.items():
        assert _fn(k) == v


def test_remapping_patch_repeated():
    remapping = {"type": "{type}{number}"}
    patch = {"type": {"cf0": "x"}}
    r = build_remapping(remapping, patch)
    r = build_remapping(r)
    _fn = r(fn)

    assert _fn("type") == "x"


@pytest.mark.parametrize(
    "remapping,patch,expected_values",
    [
        (
            {"type": "{type}{number}"},
            None,
            {"type": ("cf0", ("cf", "0")), "number": 0, "name": "t2m"},
        ),
        (
            {"my_type": "{type}{number}"},
            None,
            {"my_type": ("cf0", ("cf", "0")), "type": "cf", "number": 0, "name": "t2m"},
        ),
        (
            {"type": "{type}{number}"},
            {"type": {"cf0": "pf"}, "number": 12, "name": None},
            {"type": ("pf", ("cf", "0")), "number": 12, "name": None},
        ),
        (
            {"my_type": "{type}{number}"},
            {"my_type": {"cf0": "pf"}, "number": 12, "name": None},
            {"my_type": ("pf", ("cf", "0")), "type": "cf", "number": 12, "name": None},
        ),
        (
            {"my_type": "{type}{number}"},
            {"my_type": lambda x: x + "_", "number": 12, "name": None},
            {"my_type": ("cf0_", ("cf", "0")), "type": "cf", "number": 12, "name": None},
        ),
    ],
)
def test_remapping_collector(remapping, patch, expected_values):
    r = build_remapping(remapping, patch)
    _fn = r(fn, joiner=CollectorJoiner)
    for k, v in expected_values.items():
        assert _fn(k) == v


@pytest.mark.parametrize(
    "remapping,patch,expected_values",
    [
        (
            {"type": "{type}{number}"},
            None,
            {"type": ["type", "number"], "number": [], "name": []},
        ),
        (
            {"my_type": "{type}{number}"},
            None,
            {"my_type": ["type", "number"], "type": [], "number": [], "name": []},
        ),
        (
            {"type": "{type}{number}"},
            {"type": {"cf0": "pf"}, "number": 12, "name": None},
            {"type": ["type", "number"], "number": [], "name": []},
        ),
        (
            {"my_type": "{type}{number}"},
            {"my_type": {"cf0": "pf"}, "number": 12, "name": None},
            {"my_type": ["type", "number"], "type": [], "number": [], "name": []},
        ),
    ],
)
def test_remapping_components(remapping, patch, expected_values):
    r = build_remapping(remapping, patch)
    for k, v in expected_values.items():
        assert r.components(k) == v



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.utils.testing import get_array_backend

from earthkit.data import from_source
from earthkit.data.testing import ARRAY_BACKENDS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file


def load_array_fieldlist(path, array_backend, **kwargs):
    ds = from_source("file", path, **kwargs)
    return ds.to_fieldlist(array_backend=array_backend)
    # return FieldList.from_array(
    #     ds.values, [m.override(generatingProcessIdentifier=120) for m in ds.metadata()]
    # )


def load_grib_data(filename, fl_type, folder="example", **kwargs):
    if isinstance(filename, str):
        filename = [filename]

    if folder == "example":
        path = [earthkit_examples_file(name) for name in filename]
    elif folder == "data":
        path = [earthkit_test_data_file(name) for name in filename]
    else:
        raise ValueError(f"Invalid folder={folder}")

    if fl_type == "file":
        return from_source("file", path, **kwargs), get_array_backend("numpy")
    elif fl_type == "array":
        return load_array_fieldlist(path, "numpy", **kwargs), get_array_backend("numpy")
    elif fl_type == "memory":
        assert len(path) == 1
        with open(path[0], "rb") as f:
            ds = from_source("stream", f, read_all=True, **kwargs)
            len(ds)  # force reading
            return ds, get_array_backend("numpy")
    elif array_backend := get_array_backend(fl_type):
        return load_array_fieldlist(path, array_backend, **kwargs), array_backend
    else:
        raise ValueError(f"Invalid fl_type={fl_type}")


FL_TYPES = ["file"]
FL_TYPES.extend([b.name for b in ARRAY_BACKENDS])
FL_ARRAYS = [b.name for b in ARRAY_BACKENDS]
FL_NUMPY = ["file", "numpy"]
FL_FILE = ["file"]



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
t = ekd.create_target(
    "file",
    "res.grib",
    metadata={"shortName": "2t", "generatingProcessIdentifier": 255},
)
for field in ds:
    t.write(field)

t.close()

# case 2 - equivalent solutions
t = ekd.create_target("file", "res.grib")
for field in ds:
    t.write(data=field, shortName="2t", generatingProcessIdentifier=255)
t.close()

t = ekd.create_target("file", "res.grib")
for field in ds:
    t.write(shortName="2t", template=field, generatingProcessIdentifier=255)
t.close()

# case 3
with ekd.create_target("file", "res.grib") as t:
    for field in ds:
        t.write(field, shortName="2t", generatingProcessIdentifier=255)

# case 4
with ekd.create_target("file", "res.grib") as t:
    for field in ds:
        values = field.values + 1
        t.write(data=field, values=values, shortName="2t")
